__author__ = 'nthmost'

from ..pubmedfetcher import PubMedFetcher
from ..convert import doi2pmid
from ..exceptions import MetaPubError

from .dances import *


PUBMED_SWITCHBOARD = {
    'jstage':   {'journals': jstage_journals, 'dance': the_jstage_dive,     },
    'springer': {'journals': springer_journals, 'dance': the_springer_shag, },
    'wiley':    {'journals': wiley_journals, 'dance': the_wiley_shuffle,    },
    'jama':     {'journals': jama_journals, 'dance': the_jama_dance,        },
    'aaas':     {'journals': aaas_journals, 'dance': the_aaas_tango,        },
    'spandidos': {'journals': spandidos_journals, 'dance': the_spandidos_lambada, },
    'jci':      {'journals': jci_journals, 'dance': the_jci_jig,            },
    'scielo':   {'journals': scielo_journals, 'dance': the_scielo_chula,    },
    'najms':    {'journals': najms_journals, 'dance': the_najms_mazurka,    },
    'biomchemsoc': {'journals': biochemsoc_journals, 'dance': the_biochemsoc_saunter,   },
    'nature':   {'journals': nature_journals, 'dance': the_nature_ballet,   },
    'cell':     {'journals': cell_journals, 'dance': the_cell_pogo,         },
    'lancet':   {'journals': lancet_journals, 'dance': the_lancet_tango,    },
    'sciencedirect': {'journals': sciencedirect_journals, 'dance': the_sciencedirect_disco, },
    'karger':   {'journals': karger_journals, 'dance': the_karger_conga,    },
    'wolterskluwer': {'journals': wolterskluwer_journals, 'dance': the_wolterskluwer_volta, },
}


""" findit/logic.py

        The get_pdf_from_pma function selects possible PDF links for the
        given article represented in a PubMedArticle object.

        These links are built (not crawled) by selecting a likely-to-work URL
        pattern based on the NLM journal name abbreviation taken from the
        PubMedArticle object.

        It's recommended to use the FindIt object as the primary interface
        to this code.

        See the find_article_from_pma docstring for more info.

        *** IMPORTANT NOTE ***

        In many cases, this code performs intermediary HTTP requests in order to
        scrape a PDF url out of a page, and sometimes tests the url to make sure
        that what's being sent back is in fact a PDF.

        NO PDF DOWNLOAD IS PERFORMED; however some websites will block your IP
        address when you are performing several information lookups within a
        relatively short span of time (e.g. informa blocks if 25 HTTP connections
        are made within 5 minutes).

        If you would like these requests to go through a proxy (e.g. if you would
        like to prevent making multiple requests of the same servers, which may have
        effects like getting your IP shut off from PubMedCentral), set the
        HTTP_PROXY environment variable in your code or on the command line before
        using any FindIt functionality.
"""

def find_article_from_pma(pma, verify=True, use_nih=False):
    """ The real workhorse of FindIt.

        Based on the contents of the supplied PubMedArticle object, this function
        returns the best possible download link for a Pubmed PDF.

        Be aware that this function no longer performs doi lookups; if you want
        this handled for you, use the FindIt object (which will also record the
        doi score from the lookup for you).

        Returns (url, reason) -- url being self-explanatory, and "reason" containing
        any qualifying message about why the url came back the way it did.

        Reasons may include (but are not limited to):

            "DOI missing from PubMedArticle and CrossRef lookup failed."
            "pii missing from PubMedArticle XML"
            "No URL format for Journal %s"
            "TODO format"

        Optional params:
            use_nih      -- source PubmedCentral articles from nih.gov (NOT recommended)

        :param pma: PubMedArticle object)
        :param verify: (bool) default: True
        :param use_nih: (bool) default: False
        :return: (url, reason)
    """
    reason = ''
    url = None
    jrnl = standardize_journal_name(pma.journal)

    # === Pubmed Central: ideally we get the article from PMC if it has a PMC id.
    #
    #   Note: we're using europepmc.org rather than nih.gov (see the_pmc_twist function).
    #
    #   If we can't get the article from a PMC site, it may be that the paper is
    #   temporarily embargoed.  In that case, we may be able to fall back on retrieval
    #   from a publisher link.

    if pma.pmc:
        try:
            url = the_pmc_twist(pma, verify, use_nih)
            return (url, None)
        except MetaPubError as error:
            reason = str(error)

    # === IDENTIFIER-BASED LISTS === #

    if jrnl in simple_formats_pii.keys():
        try:
            url = the_pii_polka(pma, verify)
        except MetaPubError as error:
            reason = str(error)

    elif jrnl in simple_formats_pmid.keys():
        try:
            url = the_pmid_pogo(pma, verify)
        except MetaPubError as error:
            reason = str(error)

    elif jrnl in simple_formats_doi.keys():
        try:
            url = the_doi_slide(pma, verify)
        except MetaPubError as error:
            reason = str(error)

    elif jrnl in vip_journals.keys():
        try:
            url = the_vip_shake(pma, verify)
        except MetaPubError as error:
            reason = str(error)

    elif jrnl in vip_journals_nonstandard.keys():
        try:
            url = the_vip_nonstandard_shake(pma, verify)
        except MetaPubError as error:
            reason = str(error)

    if url:
        return (url, reason)

    # === PUBLISHER BASED LISTS === #

    # Many Biomed Central journals start with "BMC", but many more don't.
    if jrnl.find('BMC') == 0 or jrnl in BMC_journals:
        try:
            url = the_biomed_calypso(pma, verify)
        except MetaPubError as error:
            reason = str(error)

    for publisher in list(PUBMED_SWITCHBOARD.keys()):
        if jrnl in PUBMED_SWITCHBOARD[publisher]['journals']:
            try:
                url = PUBMED_SWITCHBOARD[publisher]['dance'](pma)
            except MetaPubError as error:
                reason = str(error)

    if url:
        return (url, reason)

    #if jrnl in paywall_journals:
    #    reason = 'PAYWALL: this journal has been marked in a list as "never free"'

    elif jrnl in todo_journals:
        reason = 'TODO: format example: %s' % todo_journals[jrnl]['example']

    elif jrnl in JOURNAL_CANTDO_LIST:
        reason = 'CANTDO: this journal has been marked as unsourceable'

    # aka if url is STILL None...
    if not url and not reason:
        reason = 'NOFORMAT: No URL format for journal "%s"' % jrnl

    return (url, reason)


def find_article_from_doi(doi, verify=True, use_nih=False):
    """ Pull a PubMedArticle based on CrossRef lookup (using doi2pmid),
    then run it through find_article_from_pma.

    :param doi: (string)
    :return: (url, reason)
    """
    fetch = PubMedFetcher()
    pma = fetch.article_by_pmid(doi2pmid(doi))
    return find_article_from_pma(pma, verify=verify, use_nih=use_nih)




import unicodedata

from unidecode import unidecode

from urllib.parse import urlparse, unquote

PUNCS_WE_DONT_LIKE = "[],.()<>'/?;:\"&"


def kpick(args, options, default=None):
    for opt in options:
        if args.get(opt, None):
            return args[opt]
    return default


def remove_chars(inp, chars=PUNCS_WE_DONT_LIKE, urldecode=False):
    """ Remove target characters from input string.

    :param inp: (str)
    :param chars: (str) characters to remove [default: utils.PUNCS_WE_DONT_LIKE]
    :param urldecode: (bool) whether to first urldecode the input string [default: False]
    """
    if urldecode:
        inp = unquote(inp)

    for char in chars:
        inp = inp.replace(char, '')
    return inp


def hostname_of(url):
    """ Takes a url (may or may not contain protocol prefix) and returns the simplest base form of the 
    hostname in the supplied URL.
    
    If hostname starts with 'www.', this will be stripped out.

    Examples: 
        http://www.nature.com/pr/journal/v49/n1/full/pr20018a.html --> nature.com
        https://webhome.weizmann.ac.il --> webhome.weizmann.ac.il
        https://www.ncbi.nlm.nih.gov/pubmed/17108762 --> ncbi.nlm.nih.gov

    :param url: (str)
    :return hostname: (str)
    """
    if url.startswith('http'):
        hostname = urlparse(url).hostname
    else:
        hostname = urlparse('http://' + url).hostname

    if hostname.startswith('www'):
        hostname = hostname.replace('www.', '')
    return hostname


def rootdomain_of(url):
    """ Returns the root domain of hostname of supplied URL. 

    Examples:
        http://blood.oxfordjournals.org --> oxfordjournals.org
        https://webhome.weizmann.ac.il --> ac.il
        https://regex101.com/ --> regex101.com
        https://www.ncbi.nlm.nih.gov/pubmed/17108762 --> nih.gov

    :param url: (str)
    :return rootdomain: (str)
    """
    hostname = hostname_of(url)
    return '.'.join(hostname.split('.')[-2:])


def asciify(inp):
    """ Nuke all the unicode from orbit. It's the only way to be sure.

    WARNING: this function is mostly used for Python2 compatibility and other legacy stuff,
    and may be removed in upcoming versions of metapub.

    :param inp: (str)
    :return: string converted to pure, American ASCII
    """
    # TODO: be more diplomatic than an atomic bomb: convert international chars to ascii equivalents.
    # see http://stackoverflow.com/questions/517923/what-is-the-best-way-to-remove-accents-in-a-python-unicode-string
    DeprecationWarning('asciify is a relic of The Great Python2 Unicode Wars, and will be removed in v0.6.0.')
    if inp:
        try:
            return inp.encode('ascii', 'ignore')
        except UnicodeDecodeError:
            return unicodedata.normalize('NFKD', inp.decode('utf-8')).encode('ascii', 'ignore')
    else:
        return ''


def squash_spaces(inp):
    """ Convert multiple ' ' chars to a single space.

    :param inp: (str)
    :return: same string with only one space where multiple spaces were.
    """
    return ' '.join(inp.split())


def parameterize(inp, sep='+'):
    """ Make strings suitable for submission to GET-based query service.

    Strips out the characters named in metapub.utils.PUNCS_WE_DONT_LIKE

    If inp is None, return empty string.

    :param inp: (str or None): input to be parameterized
    :param sep: (str): separator to use in place of spaces (default='+')
    :return: "parameterized" str
    """
    if inp is None:
        return ''

    inp = remove_chars(inp, PUNCS_WE_DONT_LIKE) 
    inp = squash_spaces(inp).replace(' ', sep)

    return unidecode(inp)


def deparameterize(inp, sep='+'):
    """ Somewhat-undo parameterization in string. Replace separators (sep) with spaces.

    :param inp: (str)
    :param sep: (str) default: '+'
    :return: "deparameterized" string
    """
    return inp.replace(sep, ' ')


def remove_html_markup(inp):
    """ Remove html and xml tags from text.
    Preserves HTML entities like &amp;

    :param inp: (str)
    :return: string with HTML and XML markup removed.
    """
    tag = False
    quote = False
    out = ""

    for char in inp:
        if char == '<' and not quote:
            tag = True
        elif char == '>' and not quote:
            tag = False
        elif (char == '"' or char == "'") and tag:
            quote = not quote
        elif not tag:
            out = out + char
    return out


def lowercase_keys(dct):
    """ Takes an input dictionary, returns dictionary with all keys lowercased. """
    result = {}
    for key, value in list(dct.items()):
        result[key.lower()] = value
    return result



__doc__ = "Common functions for the formatting of academic reference citations."

article_cit_fmt = '{author}. {title}. {journal}. {year}; {volume}:{pages}.{doi}'
book_cit_fmt = '{author}. {book.title}. {cdate} (Update {mdate}). In: {editors}, editors. {book.journal} (Internet). {book.book_publisher}'


def author_str(author_list_or_string, as_html=False):
    """ Helper function for constructing article citations.

    :param author_list_or_string:
    :return: author(s) str suitable for printed citation
    """
    if type(author_list_or_string) == list:
        authors = author_list_or_string
    else:
        authors = author_list_or_string.split(';')

    if len(authors) > 2:
        auth = authors[0].strip()
        if as_html:
            auth += ', <i>et al</i>'
        else:
            auth += ', et al'
    elif len(authors) == 2:
        auth = ' and '.join([aut.strip() for aut in authors])
    else:
        auth = authors[0]

    return auth


def citation(**kwargs):
    """ Returns a formatted citation string built from this article's author(s), title,
    journal, year, volume, pages, and doi.

    see cite.article and cite.book for more specific use cases.

    Note that "authors" (as list) will be used preferentially over "author" (as str).

    Keywords:
        as_html: (bool) returns citation with light HTML formatting.
        author: (str) -- prints author as-is without modification
        authors: (list) -- prints as author1 (first in list) as "Lastname_FirstInitials, et al"
        title: (str)
        journal: (str)
        year: (str or int)
        volume: (str or int)
        pages: (str) should be formatted "nn-mm", e.g. "55-58"
        doi: (str)
        
    Returns:
        citation (str)
    """
    #Author(s)
    if kwargs.get('authors', None):
        author = author_str(kwargs['authors'], as_html=kwargs.get('as_html', False))
    else:
        author = kwargs.get('author', '(None)')

    #DOI
    doi_str = '' if not kwargs.get('doi', None) else ' doi: %s' % kwargs['doi']

    #Title
    title = '(None)' if not kwargs.get('title', None) else kwargs['title'].strip('.')
    #Journal
    journal = '(None)' if not kwargs.get('journal', None) else kwargs['title'].strip('.')
    # Year
    year = '(unknown year)' if not kwargs.get('year', None) else kwargs['year']

    # Volume
    volume = '(unknown volume)' if not kwargs.get('volume', None) else str(kwargs['volume'])
    # Pages
    pages = '(unknown pages)' if not kwargs.get('pages', None) else kwargs['pages']

    #TODO: how many articles DON'T have a volume, or are missing pages? (not that important for now.)
    # article_cit_fmt = '{author}. {title}. {journal}. {year}; {volume}:{pages}.{doi}'
    return article_cit_fmt.format(author=author, volume=volume, pages=pages, year=year,
                                  title=title, journal=journal, doi=doi_str)


def article(**kwargs):
    """ Returns a formatted citation string built from this article's author(s), title,
        journal, year, volume, pages, and doi.

        This function uses the Article format citation template. For example:

        McNally EM, et al. Genetic mutations and mechanisms in dilated cardiomyopathy. Journal of Clinical Investigation. 2013; 123:19-26. doi: 10.1172/JCI62862.

        Keywords:
            journal
            title
            doi
            authors (str or list) -- if str, prints authors without modification.

        Return:
            citation (str)
    """
    return citation(**kwargs)


def book(book, **kwargs):
    """ Takes a PubMedArticle "book" and formats a citation string.  This is a special type of citation
        built mostly for NCBI GeneReviews and not currently generalizable to other academic books (yet).

        Returns a formatted citation string for a book.  A "book" needs to contain the following attributes:

            author
            title
            book_date_revised
            book_contribution_date
            editors             
            journal
            book_publisher      (may be a URL)

        This function uses the Book format citation template: 

        book_cit_fmt = '{author}. {title}. {cdate} (Update {mdate}). In: {editors}, editors. {journal} (Internet). {book_publisher}'

        For example:

        Tranebjarg L, et al. Jervell and Lange-Nielsen syndrome. 2002 Jul 29 (Updated 2014 Nov 20). In: Pagon RA, et al., editors. GeneReviews (Internet). Seattle (WA): University of Washington, Seattle; 1993-2015. Available from: https://www.ncbi.nlm.nih.gov/books/NBK1405/.

        :param book: PubMedArticle of type "book"
        :param use_html: (bool) whether to return with light HTML formatting
        :return: formatted citation string
        :rtype: str
    """
    author = author_str(book.authors_str, as_html=kwargs.get('as_html', False))

    # Special handling for GeneReviews and other books:
    if book.book_accession_id:
        mdate = book.book_date_revised.strftime('%Y %b %d')
        cdate = book.book_contribution_date.strftime('%Y %b %d')
        editors = author_str(book.book_editors, as_html=kwargs.get('as_html', False))
        if editors.endswith(', et al'):
            editors += '.'
        return book_cit_fmt.format(editors=editors, author=author, book=book,
                                   mdate=mdate, cdate=cdate)




from __future__ import absolute_import, unicode_literals

# TODO: Dustri Dance
# Dustri: see http://www.dustri.com/journals-in-english.html
dustri_journals = (
    'Clin Nephrol' 
    'Int J Clin Pharmacol Ther' 
    'Clin Neuropathol',
    )



from __future__ import absolute_import

from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

from .hostname2jrnl import HOSTNAME_TO_JOURNAL_MAP
from .urlreverse import UrlReverse




__author__ = 'nthmost'

import time
import logging
import coloredlogs

import requests

from urllib.parse import urlparse

from ..crossref import CrossRefFetcher
from ..exceptions import MetaPubError
from ..utils import asciify
from ..config import DEFAULT_CACHE_DIR
from ..pubmedfetcher import PubMedFetcher
from ..convert import doi2pmid
from ..cache_utils import get_cache_path, SQLiteCache, datetime_to_timestamp

from .logic import find_article_from_pma
from .dances import the_sciencedirect_disco, the_doi_2step, the_wiley_shuffle, the_wolterskluwer_volta

log = logging.getLogger('metapub.findit')

coloredlogs.install()

""" findit/findit.py

    Provides FindIt object, providing a tidy object layer
        into the logic.get_pdf_from_pma function. (see logic.py)

    The FindIt class allows lookups of the PDF starting from only a
    DOI or a PMID, using the following instantiation approaches:

    source = FindIt('1234567')   # assumes argument is a pubmed ID

    source = FindIt(pmid=1234567)  # pmid can be an int or a string

    source = FindIt(doi="10.xxxx/xxx.xxx")   # doi instead of pmid.

    See the FindIt docstring for more information.

    *** IMPORTANT NOTE ***

    In many cases, this code performs intermediary HTTP requests in order to
    scrape a PDF url out of a page, and sometimes tests the url to make sure
    that what's being sent back is in fact a PDF.

    If you would like these requests to go through a proxy (e.g. if you would
    like to prevent making multiple requests of the same servers, which may have
    effects like getting your IP shut off from PubMedCentral), set the
    HTTP_PROXY environment variable in your code or on the command line before
    using any FindIt functionality.
"""

CACHE_FILENAME = 'findit.db'

FINDIT_CACHE = None


pm_fetch = None

def _start_engines():
    global pm_fetch
    if not pm_fetch:
        log.debug('Started PubMedFetcher engine.')
        pm_fetch = PubMedFetcher()

def _get_findit_cache(cachedir):
    global FINDIT_CACHE
    if not FINDIT_CACHE:
        _cache_path = get_cache_path(cachedir, CACHE_FILENAME)
        FINDIT_CACHE = SQLiteCache(_cache_path)
        log.info('FindIt Cache initialized at %s', _cache_path)
    return FINDIT_CACHE


class FindIt(object):
    """ FindIt

        FindIt helps locate an article's fulltext PDF based on its pubmed ID
        or doi, using the following instantiation approaches:

        source = FindIt('1234567')   # assumes argument is a pubmed ID

        source = FindIt(pmid=1234567)  # pmid can be an int or a string

        source = FindIt(doi="10.xxxx/xxx.xxx")   # doi instead of pmid.

        The machinery in the FindIt object performs all necessary data lookups
        (e.g. looking up a missing DOI, or using a DOI to get a PubMedArticle)
        to end up with a url and reason, which attaches to the FindIt object
        in the following attributes:

        source = FindIt(pmid=PMID)
        source.url
        source.reason
        source.pmid
        source.doi
        source.doi_score

        The "doi_score" is an indication of where the DOI for this PMID ended up
        coming from. If it was supplied by the user or by PubMed, doi_score will be 100.

        If CrossRef came into play during the process to find a DOI that was missing
        for the PubMedArticle object, the doi_score will come from CrossRef (0 to 100).
    """

    def __init__(self, pmid=None, cachedir=DEFAULT_CACHE_DIR, **kwargs):

        _start_engines()

        self.pmid = pmid if pmid else kwargs.get('pmid', None)
        self.doi = kwargs.get('doi', None)
        self.url = kwargs.get('url', None)
        self.reason = None
        self.use_nih = kwargs.get('use_nih', False)
        self.use_crossref = kwargs.get('use_crossref', True)

        if self.use_crossref:
            self.crfetch = CrossRefFetcher()
            log.debug('CrossRefFetcher initialized for FindIt.')

        #TODO: revisit this whole score thing... it has changed.
        self.doi_min_score = kwargs.get('doi_min_score', 60)   #60, maybe?
        self.tmpdir = kwargs.get('tmpdir', '/tmp')
        self.doi_score = None
        self.pma = None
        self._backup_url = None

        self.verify = kwargs.get('verify', True)
        retry_errors = kwargs.get('retry_errors', False)

        if cachedir is None:
            self._cache = None
        else:
            self._cache = _get_findit_cache(cachedir)

        self._log = logging.getLogger('metapub.findit')
        if kwargs.get('debug', False):
            self._log.setLevel(logging.DEBUG)
        else:
            self._log.setLevel(logging.INFO)

        if self.pmid:
            self._load_pma_from_pmid()
        elif self.doi:
            self._load_pma_from_doi()
        else:
            raise MetaPubError(
                'Supply either a pmid or a doi to instantiate. e.g. FindIt(pmid=1234567)')

        try:
            if self._cache:
                self.url, self.reason = self.load_from_cache(verify=self.verify, retry_errors=retry_errors)
            else:
                self.url, self.reason = self.load(verify=self.verify)

        except requests.exceptions.ConnectionError as error:
            self.reason = 'TXERROR: %r' % error

    def load(self, verify=True):
        """ Interface to logic.find_article_from_pma; uses self.pma to return
        a (url, reason) tuple.

        If `verify`, actually test the link to ensure that a file could be downloaded
        from the calculated link. (Setting this to False usually speeds up FindIt considerably.)

        If url is None, reason should have a string.  If url is not None, there may
        or may not be a reason string (usually not).

        If a ConnectionError prevented lookup, returns (None, "TXERROR: <info>")

        :param verify: (bool) default: True
        :return: (url, reason) (string or None, string or None)
        """
        return find_article_from_pma(self.pma, use_nih=self.use_nih, verify=verify)

    def load_from_cache(self, verify=True, retry_errors=False):
        """ Using preloaded identifiers (self.pmid, self.doi, etc), check cache
        for article lookup results. If it's not in the cache, call self.load()
        and store the results in the cache.

        If url is None, reason should have a string.  If url is not None, there may
        or may not be a reason string (usually not).

        If self.load() comes up as a ConnectionError (usually indicating a problem
        with the internet connection), no result will be cached.

        If cache result has reason like "NOFORMAT", "TODO", "PAYWALL", or "CANTDO",
        try a fresh load; there are new formats added to FindIt all the time. :)

        :param verify: (bool) default: True
        :param retry_errors: (bool) default: False
        :return: (url, reason) (string or None, string or None)
        """
        retry_reasons = ['PAYWALL', 'TODO', 'NOFORMAT', 'CANTDO']
        if retry_errors:
            retry_reasons.append('TXERROR')

        cache_result = self._query_cache(self.pmid)
        if cache_result:
            url = cache_result['url']
            reason = '' if cache_result['reason'] is None else cache_result['reason']

            # Prefer cached results that were verified.
            # Don't return cached results in retry_reasons list above. (i.e. retry)
            if cache_result.get('verify', False) or verify == False:
                if not reason.split(':')[0] in retry_reasons:
                    return (url, reason)

        # === RETRY === #
        # we're here for one of the following reasons:
        # 1) no cache result for this query
        # 2) previous result was unverified and now verify=True
        # 3) previous result had a "reason" in retry_reasons
        url, reason = self.load(verify=verify)
        self._store_cache(self.pmid, url=url, reason=reason, verify=verify)
        return (url, reason)

    @property
    def backup_url(self):
        """ Returns a backup url to try if the first url doesn't pan out.

        (Highly experimental. Results not cached.)
        """
        if not self.doi:
            return None

        if self._backup_url is not None:
            return self._backup_url

        try:
            baseurl = the_doi_2step(self.doi)
        except MetaPubError as error:
            self._log.debug('%r', error)
            return None

        urlp = urlparse(baseurl)

        # maybe it's sciencedirect / elsevier:
        if urlp.hostname.find('sciencedirect') > -1 or urlp.hostname.find('elsevier') > -1:
            if self.pma.pii:
                try:
                    self._backup_url = the_sciencedirect_disco(self.pma)
                except Exception as error:
                    self._log.debug('%r', error)

        # maybe it's wiley:
        elif urlp.hostname.find('wiley') > -1:
            try:
                self._backup_url = the_wiley_shuffle(self.pma)
            except Exception as error:
                self._log.debug('%r', error)
                self._backup_url = None
                return None

        # maybe it's wolterskluwer:
        elif urlp.hostname.find('pt.wkhealth.com') > -1:
            try:
                self._backup_url = the_wolterskluwer_volta(self.pma)
            except Exception as error:
                self._log.debug('%r', error)
                self._backup_url = None
                return None

        # TODO maybe it's an "early" print? if so it might look like this:
        #
        # if urlp.path.find('early'):
        #    return None

        if self._backup_url is None and urlp.path.find('.') > -1:
            extension = urlp.path.split('.')[-1]
            if extension == 'long':
                self._backup_url = baseurl.replace('long', 'full.pdf')
            elif extension == 'html':
                self._backup_url = baseurl.replace(
                    'full', 'pdf').replace('html', 'pdf')

        if self._backup_url is None:
            # a shot in the dark...
            if urlp.path.endswith('/'):
                self._backup_url = baseurl + 'pdf'
            else:
                self._backup_url = baseurl + '.full.pdf'

        return self._backup_url

    def _load_pma_from_pmid(self):
        """ Loads self.pma if self.pmid is present.

        Mutates:
            self.doi (using crossref to look this information up if necessary)
            self.doi_score (100 if doi found in self.pma, else crossref score)
        """

        self.pma = pm_fetch.article_by_pmid(self.pmid)

        if self.pma.doi:
            self.doi = self.pma.doi
            self.doi_score = 100

        if self.pma.doi == None:
            if self.use_crossref:
                self._log.debug('Using CrossRef to find DOI for PMID %s', self.pmid)
                work = self.crfetch.article_by_pma(self.pma)
                if work:
                    self.doi = work.doi
                    self.doi_score = work.score
                    self._log.debug('\tFound DOI ', self.doi, ' with score ', self.doi_score)
                else:
                    self._log.debug('\tCrossRef DOI lookup failed for PMID %s.', self.pmid)
                    self.reason = 'MISSING: doi (CrossRef lookup failed)'
            else:
                self.reason = 'MISSING: doi (CrossRef lookups disabled)'

    def _load_pma_from_doi(self):
        """ Loads self.pma if self.doi is present.

        Mutates:
            self.pmid (using metapub.convert.doi2pmid)
            self.pma  (if pmid was found)
            self.doi_score (10.0 if doi found in self.pma, else crossref score)
        """
        self.pmid = doi2pmid(self.doi)
        if self.pmid:
            self.pma = pm_fetch.article_by_pmid(self.pmid)
            self.doi_score = 100
        else:
            raise MetaPubError('Could not get a pmid for doi %s' % self.doi)

    def to_dict(self):
        """ Returns a dictionary containing the public attributes of this object"""
        return {'pmid': self.pmid,
                'doi': self.doi,
                'reason': self.reason,
                'url': self.url,
                'doi_score': self.doi_score,
                }

    def _make_cache_key(self, pmid):
        """ Returns normalized key (pmid as integer) for hash lookup / store. """
        return int(pmid)

    def _store_cache(self, cache_key, **kwargs):
        """ Store supplied cache_key pointing to values supplied in kwargs.

        A time.time() timestamp will be added to the value dictionary when stored.

        There is no return from this function. Exceptions from the SQLiteCache 
        object may be raised.
        """
        cache_value = kwargs.copy()
        cache_value['timestamp'] = time.time()
        self._cache[self._make_cache_key(cache_key)] = cache_value

    def _query_cache(self, pmid, expiry_date=None):
        """ Return results of a lookup from the cache, if available.
        Return None if not available.

        Cache results are stored with a time.time() timestamp.

        When expiry_date is supplied, results from the cache past their
        sell-by date will be expunged from the cache and return will be None.

        expiry_date can be either a python datetime or a timestamp. 

        :param: cache_key: (required)
        :param: expiry_date (optional, default None)
        :rtype: (url, reason) or None
        """

        if hasattr(expiry_date, 'strftime'):
            # convert to timestamp
            sellby = datetime_to_timestamp(expiry_date)
        else:
            # make sure sellby is a number, not None
            sellby = expiry_date if expiry_date else 0

        if self._cache:
            cache_key = self._make_cache_key(pmid)
            try:
                res = self._cache[cache_key]
                timestamp = res['timestamp']
                if timestamp < sellby:
                    self._log.debug('Cache: expunging result for %s (%i)', cache_key, timestamp)
                else:
                    self._log.debug('Cache: returning result for %s (%i)', cache_key, timestamp)
                return res

            except KeyError:
                self._log.debug('Cache: no result for key %s', cache_key)
                return None
        else:
            self._log.debug('Cache disabled (self._cache is None)')
            return None




__author__ = 'nthmost'
__doc__ = '''metapub.PubMedFetcher -- tools to deal with NCBI's E-utilities interface to PubMed'''

from lxml import etree
import requests
import eutils
import logging

from .eutils_common import get_eutils_client
from .cache_utils import get_cache_path
from .pubmedarticle import PubMedArticle
from .pubmedcentral import get_pmid_for_otherid
from .pubmed_clinicalqueries import *
from .utils import kpick, parameterize, lowercase_keys, remove_chars
from .text_mining import re_pmid, is_ncbi_bookID, re_matching_quotes
from .exceptions import MetaPubError, InvalidPMID, InvalidBookID
from .base import Borg
from .config import DEFAULT_EMAIL, API_KEY

log = logging.getLogger('metapub.pubmedfetcher')

def get_uids_from_esearch_result(xmlstr):
    dom = etree.fromstring(xmlstr)
    uids = []
    idlist = dom.find('IdList')
    for item in idlist.findall('Id'):
        uids.append(item.text.strip())
    return uids

def parse_related_pmids_result(xmlstr):
    outd = {}
    dom = etree.fromstring(xmlstr)
    for linkset in dom.findall('LinkSet/LinkSetDb'):
        heading = linkset.find('LinkName').text.split('_')[-1]
        outd[heading] = []
        for Id in linkset.findall('Link/Id'):
            outd[heading].append(Id.text)
    return outd

class PubMedFetcher(Borg):
    '''PubMedFetcher (a Borg singleton object backed by an optional SQLite cache)

    An interaction layer for querying via specified method to return PubMedArticle objects.

    Currently available methods: eutils

    Basic Usage:

        fetch = PubMedFetcher()

    To specify a service method (more coming soon):

        fetch = PubMedFetcher('eutils')

    To return an article by querying the service with a known PMID or NCBI Book ID:

        paper = fetch.article_by_pmid('123456')
        book = fetch.article_by_pmid('NBK1234')

    Similar methods exist for returning papers by DOI and PM Central id:

        paper = fetch.article_by_doi('10.1038/ng.379')
        paper = fetch.article_by_pmcid('PMC3458974')

    Finally, you can search for PMIDs via citation details by using the pmids_for_citation
    method, for which you usually only need 3 out of 5 details to triangulate on a good result.

        pmids = fetch.pmids_for_citation(journal='Science', year='2008', volume='4',
                first_page='7', author_name='Grant')
    '''

    _cache_filename = 'pubmedfetcher.db'

    def __init__(self, method='eutils', cachedir='default'):
        self.method = method

        if method=='eutils':
            self._cache_path = get_cache_path(cachedir, self._cache_filename)
            self.qs = get_eutils_client(self._cache_path)
            self.article_by_pmid = self._eutils_article_by_pmid
            self.article_by_pmcid = self._eutils_article_by_pmcid
            self.article_by_doi = self._eutils_article_by_doi
            self.pmids_for_query = self._eutils_pmids_for_query
        else:
            raise NotImplementedError('coming soon: fetch from local pubmed via medgen-mysql or filesystem cache.')

    def _eutils_article_by_pmid(self, pmid):
        pmid = str(pmid)
        try:
            result = self.qs.efetch({'db': 'pubmed', 'id': pmid})
        except eutils.EutilsRequestError:
            raise MetaPubError('Invalid ID "%s" (rejected by Eutils); please check the number and try again.' % pmid)

        if result is None:
            return None

        #if result.find('ERROR') > -1:
        #    raise MetaPubError('PMID %s returned ERROR; cannot construct PubMedArticle' % pmid)

        pma = PubMedArticle(result)
        if pma.pmid is None:
            raise InvalidPMID('Pubmed ID "%s" not found' % pmid)
        return pma

    def _eutils_article_by_pmcid(self, pmcid):
        # if user submitted a bare number, prepend "PMC" to make sure it is submitted correctly
        # the conversion API at pubmedcentral.
        pmcid = str(pmcid)
        # if this was a flat number (use re_pmid to cheat this), try prepending "PMC" to it.
        if re_pmid.findall(pmcid)[0] == pmcid:
            pmcid = 'PMC'+pmcid

        pmid = get_pmid_for_otherid(pmcid)
        if pmid is None:
            raise MetaPubError('No PMID available for PubMedCentral id %s' % pmcid)
        return self._eutils_article_by_pmid(pmid)

    def _eutils_article_by_doi(self, doi):
        pmid = get_pmid_for_otherid(doi)
        if pmid is None:
            raise MetaPubError('No PMID available for doi %s' % doi)
        return self._eutils_article_by_pmid(pmid)

    def _eutils_pmids_for_query(self, query='', since=None, until=None, retstart=0, retmax=250,
                pmc_only=False, **kwargs):
        '''returns list of pmids for given freeform query string plus keyword arguments.

        Freeform queries encased in quotes will be considered "exact match" queries.

        All Pubmed Advanced Query tokens (e.g. "TI" for title) are supported, plus many
        "soft" keywords that will, when parsed, map to the correct token for you. For example,
        you can supply journal name with any of the following keywords:

            "journal" == "jtitle" == "journal_title" == "TA"

        Keyword arguments are case-INsensitive since they are lowercased upon arrival.

        Example of series of queries to accomplish "pagination" of data:

        first_250 = fetch.pmids_for_query('some query')
        second_250 = fetch.pmids_for_query('some query', retstart=500, retmax=250)

        :param: query (string) default ''
        :param: since (string) default None  # Y/m/d format expected. Y alone or Y/m allowed.
        :param: until (string) default None  # Y/m/d format expected. Y alone or Y/m allowed.
        :param: retstart (int) default 0
        :param: retmax (int) default 250
        :param: pmc_only (bool) default False  # constructs query to only search Pubmed Central.
        '''

        # lowercase all the things.
        kwargs = lowercase_keys(kwargs)

        q = {}

        query = query.strip()
        # if we find brackets in the query string, assume they are query keyword tags.
        # otherwise, submit the query string with an "ALL" keyword tag.
        if query.find('[') == -1 and not kwargs.get('clinical_query', False):
            # if this query is surrounded in quotation marks, consider it an "exact match"
            # search against "ALL" fields. Otherwise, leave it untouched.
            # Useful to keep in a list in you run into other unicode quotes being used
            quotes = ['\"', '\'']
            if query and query[0] in quotes:
                # Allow for multiple quoted terms
                matches = [x.group(1) for x in re_matching_quotes.finditer(query)]
                for m in matches:
                    query += ' "%s"[ALL]' % m

        # Search within date range (since / until)
        #
        # working examples. search by creation date only works within defined ranges (not "< X" or "> Y")
        # ("2015/3/1"[Date - Create] : "2015/3/3"[Date - Create])
        # ("2015/2/14"[CRDT] : "2015/3/14"[CRDT])
        created_date_template = '"%s"[CRDT]'
        date_range_template = " (%s : %s)"
        if since:
            start = created_date_template % since
            if until:
                end = created_date_template % until
            else:
                end = '"3000"[CRDT]'

            query += date_range_template % (start, end)

        # unique ID referents.
        q['PMID'] = kpick(kwargs, options=['pmid', 'uid', 'pubmed_id'])
        q['AID'] = kpick(kwargs, options=['aid', 'doi'])
        q['book'] = kwargs.get('book', None)
        q['JID'] = kpick(kwargs, options=['jid', 'nlm uid', 'nlm unique id'])
        q['ISBN'] = kwargs.get('ISBN', None)
        q['RN'] = kpick(kwargs, options=['rn', 'rcn', 'ecn'])
        q['GR'] = kpick(kwargs, options=['gr', 'grant number'])

        # Pubmed Date features:
        q['DA'] = kpick(kwargs, options=['da', 'date created'])
        q['LR'] = kpick(kwargs, options=['lr', 'date revised', 'date last revised'])
        q['EDAT'] = kpick(kwargs, options=['edat', 'entrez date'])

        # Journal name:
        q['TA'] = kpick(kwargs, options=['ta', 'journal', 'jtitle', 'journal_title'])

        # Article-level characteristics (title, authors, etc):
        q['TIAB'] = kpick(kwargs, options=['tiab', 'abstract', 'title/abstract'])
        q['TI'] = kpick(kwargs, options=['ti', 'title', 'atitle', 'article_title'])
        q['TT'] = kpick(kwargs, options=['tt', 'transliterated title'])

        q['AU'] = kpick(kwargs, options=['au', 'author'])
        q['1AU'] = kpick(kwargs, options=['1au', 'aulast', 'author1_lastfm', 'author1_last_fm'])
        q['FAU'] = kpick(kwargs, options=['fau', 'first_author', 'author1'])
        q['LASTAU'] = kpick(kwargs, options=['lastau', 'last author'])
        q['CN'] = kpick(kwargs, options=['cn', 'corporate author'])
        q['FIR'] = kpick(kwargs, options=['fir', 'full investigator name'])
        q['IR'] = kpick(kwargs, options=['ir', 'investigator'])
        q['PG'] = kpick(kwargs, options=['pg', 'pages', 'spage', 'first_page'])

        # Volume / Issue characteristics
        q['IP'] = kpick(kwargs, options=['ip', 'issue'])
        q['VTI'] = kpick(kwargs, options=['vta', 'volume title'])
        q['VI'] = kpick(kwargs, options=['vi', 'volume', 'vol'])

        # Content characteristics
        q['LA'] = kpick(kwargs, options=['la', 'language'])
        q['TW'] = kpick(kwargs, options=['tw', 'text'])
        q['PS'] = kpick(kwargs, options=['ps', 'personal name as subject'])
        q['PA'] = kpick(kwargs, options=['pa', 'pharmacological action'])
        q['SB'] = kpick(kwargs, options=['sb', 'subset'])
        q['NM'] = kpick(kwargs, options=['nm', 'supplementary concept'])

        # MeSH characteristics
        q['MHDA'] = kpick(kwargs, options=['mhda', 'mesh date'])
        q['MH'] = kpick(kwargs, options=['mh', 'mesh', 'mesh terms'])
        q['MAJR'] = kpick(kwargs, options=['majr', 'mesh major topic', 'mesh major'])
        q['SH'] = kpick(kwargs, options=['sh', 'mesh subheadings'])

        # Publication characteristics
        q['DCOM'] = kpick(kwargs, options=['dcom', 'completion date'])
        q['DP'] = kpick(kwargs, options=['dp', 'date of publication', 'year', 'pdat']) #most aligned w/ PubMedArticle.year and CrossRef 'year'
        q['LID'] = kpick(kwargs, options=['lid', 'location id', 'location identifier'])
        q['PUBN'] = kpick(kwargs, options=['pubn', 'publisher'])
        q['PT'] = kpick(kwargs, options=['pt', 'pubmed_type', 'publication type'])
        q['PL'] = kpick(kwargs, options=['pl', 'place of publication'])

        # Miscellaneous, alphabetized by Medline feature tag.
        q['AD'] = kpick(kwargs, options=['ad', 'affiliation'])
        q['OT'] = kpick(kwargs, options=['ot', 'other term'])
        q['NM'] = kpick(kwargs, options=['nm', 'substance name'])
        q['SI'] = kpick(kwargs, options=['si', 'secondary source id'])

        for feature in q.keys():
            if q[feature] != None:
                query += ' "%s"[%s]' % (q[feature], feature)

        # option to query pubmed central only:
        # pubmed pmc[sb]
        if pmc_only:
            query += ' "pubmed pmc"[sb]'

        log.debug('pmids_for_query: querying %s', query)

        result = self.qs.esearch({ 'db': 'pubmed', 'term': query,
                                   'retmax': retmax, 'retstart': retstart })
        return get_uids_from_esearch_result(result)

    def pmids_for_clinical_query(self, query, category, optimization='broad',
            since=None, until=None, retstart=0, retmax=250, pmc_only=False, **kwargs):
        '''Takes a query and a category (required, see below) and returns a list
        of pubmed IDs returned by NCBI for that query.

        See also PubMedFetcher.pmids_for_query for other parameters.

            available categories:

                therapy
                diagnosis
                etiology
                prognosis
                prediction

           available optimizations:
                broad   (default)
                narrow

        :param: query (string)
        :param: category (string)
        :param: optimization (string) [default: broad]
        :return: list of pubmed IDs
        '''

        key = '%s_%s' % (category, optimization)
        if query:
            query += ' '+ clinical_query_map[key]
        else:
            raise MetaPubError('Query string required for Clinical Query.')

        kwargs['clinical_query'] = True
        return self.pmids_for_query(query, retstart=retstart, retmax=retmax, since=since, until=until, **kwargs)


    def pmids_for_medical_genetics_query(self, query, category='all', since=None, until=None,
                    retstart=0, retmax=250, pmc_only=False, **kwargs):
        '''Takes a query and category (see below) and returns a list of pubmed IDs.
        IDs returned by NCBI for that query.

        See also PubMedFetcher.pmids_for_query for other parameters.

            available categories:

                all     (default)
                diagnosis
                differential_diagnosis
                clinical_description
                management
                genetic_counseling
                genetic_testing

        :param: query (string)
        :param: category (string) [default: all]
        :return: list of pubmed IDs
        '''
        if query:
            query += ' '+ medical_genetics_query_map[category]
        else:
            raise MetaPubError('Query string required for Medical Genetics query.')

        kwargs['clinical_query'] = True
        return self.pmids_for_query(query, retstart=retstart, retmax=retmax, since=since, until=until, **kwargs)

    def pmids_for_citation(self, **kwargs):
        '''returns list of pmids for given citation. requires at least 3/5 of these keyword arguments:
            jtitle or journal (journal title)
            year or date
            volume
            spage or first_page (starting page / first page)
            aulast (first author's last name) or author1_first_lastfm (as produced by PubMedArticle class)

        Strings submitted for journal/jtitle will be run through metapub.utils.remove_chars to deal with HTML-
        encoded characters and to remove punctuation.
        '''
        # output format in return:
        # journal_title|year|volume|first_page|author_name|your_key|
        base_uri = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/ecitmatch.cgi?db=pubmed&retmode=xml&bdata={journal_title}|{year}|{volume}|{first_page}|{author_name}|{api_key}|'

        kwargs = lowercase_keys(kwargs)
        journal_title = remove_chars(kpick(kwargs, options=['jtitle', 'journal', 'journal_title'], default=''), urldecode=True)
        author_name = _reduce_author_string(kpick(kwargs,
                        options=['aulast', 'author1_last_fm', 'author', 'authors'], default=''))
        first_page = kpick(kwargs, options=['spage', 'first_page'], default='')
        year = kpick(kwargs, options=['year', 'date', 'pdat'], default='')
        volume = kpick(kwargs, options=['volume'], default='')

        inp_dict = { 'journal_title': parameterize(journal_title, '+'),
                     'year': str(year),
                     'volume': str(volume),
                     'first_page': str(first_page),
                     'author_name': parameterize(author_name, '+'),
                     'api_key': API_KEY or ''
                   }

        # clean up any "n/a" values.  eutils doesn't understand them.
        for k in inp_dict:
            if inp_dict[k].lower() == 'n/a':
                inp_dict[k] = ''

        req = base_uri.format(**inp_dict)
        log.debug('pmids_for_citation: querying with %s', req)

        content = requests.get(req).text
        pmids = []
        for item in content.split('\n'):
            if item.strip():
                pmid = item.split('|')[-1]
                pmids.append(pmid.strip())
        return pmids

    def related_pmids(self, pmid):
        '''For supplied pmid, return related ids of related pubmed articles,
        organized into a dictionary keyed by type of relation.  The keys include:

            * pubmed    (all related links)
            * citedin   (papers that cited this paper)
            * five      (the "five" that pubmed displays as the top related results)
            * reviews   (review papers that cite this paper)
            * combined  (?)

        query example:
        https://eutils.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi?retmode=xml&dbfrom=pubmed&id=14873513&cmd=neighbor
        '''
        outd = { }
        xmlstr = self.qs.elink( { 'dbfrom': 'pubmed', 'id': pmid, 'cmd': 'neighbor' } )
        outd = parse_related_pmids_result(xmlstr)
        return outd

    def pmid_for_bookID(self, book_id):
        '''For supplied NCBI Book ID, use the pubmed advanced query API to find its PMID.

        Not all NCBI Books have PMIDs.  If there is no associated PMID, this returns None.

        :param book_id: (str) e.g. "NBK2020"
        :return: (str) or None -- PMID if found, None otherwise.
        '''
        book_id = book_id.strip()
        if not is_ncbi_bookID(book_id):
            raise InvalidBookID('{} is not an NCBI Book ID'.format(book_id))
        res = self.pmids_for_query(book_id)
        if len(res) > 0:
            return res[0].strip()
        return None

def _reduce_author_string(author_string):
    # try splitting by commas
    authors = author_string.split(',')
    if len(authors)<2:
        # try splitting by semicolons
        authors = author_string.split(';')

    author1 = authors[0]
    # presume last name is at the end of the string
    return author1.split(' ')[-1]



"""
Search Field Descriptions and Tags

from https://www.ncbi.nlm.nih.gov/books/NBK3827/

Affiliation [AD]
Article Identifier [AID]
All Fields [ALL]
Author [AU]
Author Identifier [AUID]
Book [book]
Comment Corrections
Corporate Author [CN]
Create Date [CRDT]
Completion Date [DCOM]
EC/RN Number [RN]
Editor [ED]
Entrez Date [EDAT]
Filter [FILTER]
First Author Name [1AU]
Full Author Name [FAU]
Full Investigator Name [FIR]
Grant Number [GR]   Investigator [IR]
ISBN [ISBN]
Issue [IP]
Journal [TA]
Language [LA]
Last Author [LASTAU]
Location ID [LID]
MeSH Date [MHDA]
MeSH Major Topic [MAJR]
MeSH Subheadings [SH]
MeSH Terms [MH]
Modification Date [LR]
NLM Unique ID [JID]
Other Term [OT]
Owner
Pagination [PG]
Personal Name as Subject [PS]   Pharmacological Action [PA]
Place of Publication [PL]
PMID [PMID]
Publisher [PUBN]
Publication Date [DP]
Publication Type [PT]
Secondary Source ID [SI]
Subset [SB]
Supplementary Concept[NM]
Text Words [TW]
Title [TI]
Title/Abstract [TIAB]
Transliterated Title [TT]
UID [PMID]
Version
Volume [VI]
"""

'''<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE eSearchResult PUBLIC "-//NLM//DTD esearch 20060628//EN" "https://eutils.ncbi.nlm.nih.gov/eutils/dtd/20060628/esearch.dtd">
<eSearchResult><Count>1</Count><RetMax>1</RetMax><RetStart>0</RetStart><QueryKey>1</QueryKey><WebEnv>NCID_1_129952677_165.112.9.37_9001_1426564126_1266564592_0MetA0_S_MegaStore_F_1</WebEnv><IdList>
<Id>25023161</Id>
</IdList><TranslationSet><Translation>     <From>Journal of Neural Transmission[TA]</From>     <To>"J Neural Transm"[Journal] OR "J Neural Transm"[Journal] OR "J Neural Transm Suppl"[Journal] OR "J Neural Transm Park Dis Dement Sect"[Journal] OR "J Neural Transm Gen Sect"[Journal]</To>    </Translation></TranslationSet><TranslationStack>   <TermSet>    <Term>2014[DP]</Term>    <Field>DP</Field>    <Count>1171381</Count>    <Explode>N</Explode>   </TermSet>   <TermSet>    <Term>"J Neural Transm"[Journal]</Term>    <Field>Journal</Field>    <Count>1177</Count>    <Explode>N</Explode>   </TermSet>   <TermSet>    <Term>"J Neural Transm"[Journal]</Term>    <Field>Journal</Field>    <Count>3005</Count>    <Explode>N</Explode>   </TermSet>   <OP>OR</OP>   <TermSet>    <Term>"J Neural Transm Suppl"[Journal]</Term>    <Field>Journal</Field>    <Count>1409</Count>    <Explode>N</Explode>   </TermSet>   <OP>OR</OP>   <TermSet>    <Term>"J Neural Transm Park Dis Dement Sect"[Journal]</Term>    <Field>Journal</Field>    <Count>226</Count>    <Explode>N</Explode>   </TermSet>   <OP>OR</OP>   <TermSet>    <Term>"J Neural Transm Gen Sect"[Journal]</Term>    <Field>Journal</Field>    <Count>526</Count>    <Explode>N</Explode>   </TermSet>   <OP>OR</OP>   <OP>GROUP</OP>   <OP>AND</OP>   <TermSet>    <Term>121[VI]</Term>    <Field>VI</Field>    <Count>44530</Count>    <Explode>N</Explode>   </TermSet>   <OP>AND</OP>   <TermSet>    <Term>Freitag[1AU]</Term>    <Field>1AU</Field>    <Count>496</Count>    <Explode>N</Explode>   </TermSet>   <OP>AND</OP>  </TranslationStack><QueryTranslation>2014[DP] AND ("J Neural Transm"[Journal] OR "J Neural Transm"[Journal] OR "J Neural Transm Suppl"[Journal] OR "J Neural Transm Park Dis Dement Sect"[Journal] OR "J Neural Transm Gen Sect"[Journal]) AND 121[VI] AND Freitag[1AU]</QueryTranslation></eSearchResult>'''



import os
import unittest

from metapub.eutils_common import get_eutils_client

CACHE_PATH = "test_cache_eutils_client.sqlite"

class TestGetEutilsClient(unittest.TestCase):

    def tearDown(self) -> None:
        if os.path.exists(CACHE_PATH):
            os.remove(CACHE_PATH)
        return super().tearDown()

    def test_get_eutils_client(self):
        client = get_eutils_client(CACHE_PATH)
        assert client._cache is not None

    def test_get_eutils_client_no_caching(self):
        client = get_eutils_client(None)
        assert client._cache is None



from __future__ import absolute_import, unicode_literals

HOSTNAME_TO_JOURNAL_MAP = {
		'molbiolcell.org': 'Mol Biol Cell',
		'biomedgerontology.oxfordjournals.org': 'J Gerontol A Biol Sci Med Sci',
		'jrr.oxfordjournals.org': 'Radiat Res',
		'ajpendo.physiology.org': 'Am J Physiol Endocrinol Metab',
		'cancerdiscovery.aacrjournals.org': 'Cancer Discov',
		'ajpheart.physiology.org': 'Am J Physiol Heart Circ Physiol',
		'jnnp.bmj.com': 'J Neurol Neurosurg Psychiatry',
		'jvi.asm.org': 'J Virol',
		'jisakos.bmj.com': 'J ISAKOS',
		'jme.bmj.com': 'J Med Ethics',
		'adc.bmj.com': 'Arch Dis Child',
		'atvb.ahajournals.org': 'Arterioscler Thromb Vasc Biol',
		'erc.endocrinology-journals.org': 'Endocr Relat Cancer',
		'qjmed.oxfordjournals.org': 'QJM',
		'ajcp.ascpjournals.org': 'Am J Clin Pathol',
		'jid.oxfordjournals.org': 'J Infect Dis',
		'jfprhc.bmj.com': 'J Fam Plann Reprod Health Care',
		'ang.sagepub.com': 'Angiology',
		'jcn.sagepub.com': 'J Child Neurol',
		'iovs.org': 'Invest Ophthalmol Vis Sci',
		'iovs.org': 'IOVS',
		'ndt.oxfordjournals.org': 'Nephrol Dial Transplant',
		'plantcell.org': 'Plant Cell',
		'erj.ersjournals.com': 'Eur Respir J',
		'injuryprevention.bmj.com': 'Inj Prev',
		'genesdev.cshlp.org': 'Genes Dev',
		'vetrecordopen.bmj.com': 'Vet Rec Open',
		'vetrecordcasereports.bmj.com': 'Vet Rec Case Rep',
		'fasebj.org': 'FASEB J',
		'ajh.oxfordjournals.org': 'Am J Hypertens',
		'jco.ascopubs.org': 'J Clin Oncol',
		'mct.aacrjournals.org': 'Mol Cancer Ther',
		'hyper.ahajournals.org': 'Hypertension',
		'cat.sagepub.com': 'Clin Appl Thromb Hemost',
		'rheumatology.oxfordjournals.org': 'Rheumatology (Oxford)',
		'jem.rupress.org': 'J Exp Med',
		'jmg.bmj.com': 'J Med Genet',
		'ebmh.bmj.com': 'Evid Based Mental Health',
		'ajpgi.physiology.org': 'Am J Physiol Gastrointest Liver Physiol',
		'care.diabetesjournals.org': 'Diabetes Care',
		'dmd.aspetjournals.org': 'Drug Metab Dispos',
		'ard.bmj.com': 'Ann Rheum Dis',
		'jap.physiology.org': 'J Appl Physiol',
		'aac.asm.org': 'Antimicrob Agents Chemother',
		'cfp.ca': 'Can Fam Physician',
		'clincancerres.aacrjournals.org': 'Clin Cancer Res',
		'bmjopenrespres.bmj.com': 'BMJ Open Resp Res',
		'diabetes.diabetesjournals.org': 'Diabetes',
		'jramc.bmj.com': 'J R Army Med Corps',
		'openheart.bmj.com': 'Open Heart',
		'ajpcell.physiology.org': 'Am J Physiol Cell Physiol',
		'physiologyonline.physiology.org': 'Physiology (Bethesda)',
		'gut.bmj.com': 'Gut',
		'mct.aacrjournals.org': 'Mol Canc Therapeut',
		'cpj.sagepub.com': 'Clin Pediatr',
		'humrep.oxfordjournals.org': 'Hum Reprod',
		'pcp.oxfordjournals.org': 'Plant Cell Physiol',
		'peds.oxfordjournals.org': 'Protein Eng',
		'jbjs.org': 'J Bone Joint Surg Am',
		'bjo.bmj.com': 'BMC Ophthalmol',
		'ajplung.physiology.org': 'Am J Physiol Lung Cell Mol Physiol',
		'jdr.sagepub.com': 'J Dent Res',
		'physreports.physiology.org': 'Physiol Rep',
		'physrev.physiology.org': 'Physiol Rev',
		'asm.sagepub.com': 'Assessment',
		'jac.oxfordjournals.org': 'J Antimicrob Chemother',
		'clinchem.org': 'Clin Chem',
		'jah.sagepub.com': 'J Aging Health',
		'cid.oxfordjournals.org': 'Clin Infect Dis',
		'msphere.asm.org': 'mSphere',
		'molehr.oxfordjournals.org': 'Mol Hum Reprod',
		'lup.sagepub.com': 'Lupus',
		'mh.bmj.com': 'Med Humanities',
		'aim.bmj.com': 'Acupunct Med',
		'ec.asm.org': 'Eukaryot Cell',
		'bmjopensem.bmj.com': 'BMJ Open Sport Exerc Med',
		'ajprenal.physiology.org': 'Am J Physiol Renal Physiol',
		'acb.sagepub.com': 'Ann Clin Biochem',
		'qir.bmj.com': 'BMJ Qual Improv Report',
		'stroke.ahajournals.org': 'Stroke',
		'emboj.embopress.org': 'EMBO J',
		'jcm.asm.org': 'J Clin Microbiol',
		'ajpregu.physiology.org': 'Am J Physiol Regul Integr Comp Physiol',
		'circep.ahajournals.org': 'Circ Arrhythm Electrophysiol',
		'drc.bmj.com': 'BMJ Open Diabetes Res Care',
		'jim.bmj.com': 'J Investig Med',
		'jnci.oxfordjournals.org': 'J Natl Cancer Inst',
		'bja.oxfordjournals.org': 'Br J Anaesth',
		'eje-online.org': 'Eur J Endocrinol',
		'ebn.bmj.com': 'Evid Based Nurs',
		'ar.iiarjournals.org': 'Anticancer Res',
		'iai.asm.org': 'Infect Immun',
		'inpractice.bmj.com': 'In Pract',
		'dtb.bmj.com': 'Drug Ther Bull',
		'eolj.bmj.com': 'End Life J',
		'pediatrics.aappublications.org': 'Pediatrics',
		'physiolgenomics.physiology.org': 'Physiol Genomics',
		'genomebiology.com': 'Genome Biol',
		'jimmunol.org': 'J Immunol',
		'fn.bmj.com': 'Arch Dis Child Fetal Neonatal Ed',
		'innovations.bmj.com': 'BMJ Innov',
		'heart.bmj.com': 'Heart',
		'emj.bmj.com': 'Emerg Med J',
		'ee.oxfordjournals.org': 'Environ Entomol',
		'mbio.asm.org': 'MBio',
		'mcb.asm.org': 'Mol Cell Biol',
		'pnas.org': 'Proc Natl Acad Sci USA',
		'aem.asm.org': 'Appl Environ Microbiol',
		'qualitysafety.bmj.com': 'BMJ Qual Saf',
		'stel.bmj.com': 'BMJ STEL',
		'jpet.aspetjournals.org': 'J Pharmacol Exp Ther',
		'jra.sagepub.com': 'J Renin Angiotensin Aldosterone Syst',
		'breast-cancer-research.com': 'Breast Cancer Res',
		'eurheartj.oxfordjournals.org': 'Eur Heart J',
		'genome.cshlp.org': 'Genome Res',
		'jb.asm.org': 'J Bacteriol',
		'jcb.rupress.org': 'J Cell Biol',
		'mutage.oxfordjournals.org': 'Mutagenesis',
		'haematologica.org': 'Haematologica',
		'jcp.bmj.com': 'J Clin Pathol',
		'ojrd.com': 'Orphanet J Rare Dis',
		'veterinaryrecord.bmj.com': 'Vet Rec',
		'femsle.oxfordjournals.org': 'FEMS Microbiol Lett',
		'jneurosci.org': 'J Neurosci',
		'jhl.sagepub.com': 'J Hum Lact',
		'oem.bmj.com': 'Occup Environ Med',
		'tobaccocontrol.bmj.com': 'Tob Control',
		'cebp.aacrjournals.org': 'Cancer Epidemiol Biomarkers Prev',
		'jme.endocrinology-journals.org': 'J Mol Endocrinol',
		'genomea.asm.org': 'Genome Announc',
		'spcare.bmj.com': 'BMJ Support Palliat Care',
		'pmj.bmj.com': 'Postgrad Med J',
		'molpharm.aspetjournals.org': 'Mol Pharmacol',
		'jbc.org': 'J Biol Chem',
		'peds.oxfordjournals.org': 'Protein Eng Des Sel',
		'sti.bmj.com': 'Sex Transm Infect',
		'jasn.asnjournals.org': 'J Am Soc Nephrol',
		'pn.bmj.com': 'Pract Neurol',
		'msystems.asm.org': 'mSystems',
		'thorax.bmj.com': 'Thorax',
		'fg.bmj.com': 'Frontline Gastroenterol',
		'cmr.asm.org': 'Clin Microbiol Rev',
		'circ.ahajournals.org': 'Circulation',
		'ebm.bmj.com': 'Evid Based Med',
		'carcin.oxfordjournals.org': 'Carcinogenesis',
		'lupus.bmj.com': 'Lupus Sci Med',
		'jlr.org': 'J Lipid Res',
		'jb.physiology.org': 'J Neurophysiol',
		'jrheum.org': 'J Rheumatol',
		'circres.ahajournals.org': 'Circ Res',
		'esmoopen.bmj.com': 'ESMO Open',
		'ejhp.bmj.com': 'Eur J Hosp Pharm',
		'neurology.org': 'Neurology',
		'ep.bmj.com': 'Arch Dis Child Educ Pract Ed',
		'cancerres.aacrjournals.org': 'Cancer Res',
		'bmjopen.bmj.com': 'BMJ Open',
		'annonc.oxfordjournals.org': 'Ann Oncol',
		'jn.nutrition.org': 'J Nutr',
		'cardiovascres.oxfordjournals.org': 'Cardiovasc Res',
		'heartasia.bmj.com': 'Heart Asia',
		'circgenetics.ahajournals.org': 'Circ Cardiovasc Genet',
		'cvi.asm.org': 'Clin Vaccine Immunol',
		'sciencemag.org': 'Science',
		'jnis.bmj.com': 'J Neurointerv Surg',
		'jcs.biologists.org': 'J Cell Sci',
		'hmg.oxfordjournals.org': 'Hum Mol Genet',
		'mmbr.asm.org': 'Microbiol Mol Biol Rev',
		'rmdopen.bmj.com': 'RMD Open',
		'nar.oxfordjournals.org': 'Nucleic Acids Res',
		'bjsm.bmj.com': 'Br J Sports Med',
		'jech.bmj.com': 'J Epidemiol Community Health',
		'brain.oxfordjournals.org': 'Brain',
		'cpj.sagepub.com': 'Clin Pediatr (Phila)',
		'jcem.endojournals.org': 'J Clin Endocrinol Metab',
		'joe.endocrinology-journals.org': 'J Endocrinol',
		'bmj.com': 'BMJ',
		'bloodjournal.org': 'Blood',
		'jmd.amjpathol.org': 'J Mol Diagn',
		'heartrhythmjournal.com': 'Heart Rhythm',
		'ajconline.org': 'Am J Cardiol',
		'cancergeneticsjournal.org': 'Cancer Genet Cytogenet',
		'epires-journal.com': 'Epilepsy Res',
		'amjmed.com': 'Am J Med',
		'nmd-journal.com': 'Neuromuscul Disord',
		'arcmedres.com': 'Arch Med Res',
		'jpeds.com': 'J Pediatr',
		'clineu-journal.com': 'Clin Neurol Neurosurg',
		'ijporlonline.com': 'Int J Pediatr Otorhinolaryngol',
		'mgmjournal.com': 'Mol Genet Metab',
		'ahjonline.com': 'Am Heart J',
		'cancerci.com': 'Cancer Cell Int',
		'internationaljournalofcardiology.com': 'Int J Cardiol',
		'akademiai.com': 'Orv Hetil',
		'placentajournal.org': 'Placenta',
		'prd-journal.com': 'Parkinsonism Relat Disord',
		'pedneur.com': 'Pediatr Neurol',
		'cancerletters.info': 'Cancer Lett',
		'neurobiologyofaging.org': 'Neurobiol Aging',
		'metabolismjournal.com': 'Metabolism',
		'thebonejournal.com': 'Bone',
		'thrombosisresearch.com': 'Thromb Res',
		'brainanddevelopment.com': 'Brain Dev',
		'jpurol.com': 'J Pediatr Urol',
		'metabolismjournal.com': 'Metab Clin Exp',
		'worldneurosurgery.org': 'Surg Neurol',
		'jmmc-online.com': 'J Mol Cell Cardiol',
		'gastrojournal.org': 'Gastroenterology',
		'exphem.org': 'Exp Hematol',
		'medical-hypotheses.com': 'Med Hypotheses',
		'atherosclerosis-journal.com': 'Atherosclerosis',
		'gynecologiconcology-online.net': 'Gynecol Oncol',
		'aaojournal.org': 'Ophthalmology',
		'biologicalpsychiatryjournal.com': 'Biol Psychiatry',
		'ejpn-journal.com': 'Eur J Paediatr Neurol',
		'jdsjournal.com': 'J Dermatol',
		'ajo.com': 'Am J Ophthalmol',
		'fertstert.org': 'Fertil Steril',
		'jns-journal.com': 'J Neurol Sci',
		'diabetesresearchclinicalpractice.com': 'Diabetes Res Clin Pract',
		'clinsci.org': 'Clin Sci',
		'biochemj.org': 'Biochem J',
		'sciencemag.org': 'Science',
		'stm.sciencemag.org': 'Sci Transl Med',
		'advances.sciencemag.org': 'Sci Adv',
		'stke.sciencemag.org': 'Sci Signal',
		'joponline.org': 'J Periodontol',
		'medicinabuenosaires.com': 'Medicina (B Aires)',
}



import os
import coloredlogs
import logging

coloredlogs.install()

log = logging.getLogger('metapub.config')

PKGNAME = 'metapub'

# email address of this package's maintainer, included with certain NCBI calls.
# you can change this to another email address if you would prefer to be contacted.
DEFAULT_EMAIL = 'naomi@nthmost.com'

# where to place XML (temporarily) when downloaded.
TMPDIR = '/tmp'

# default cache directory for SQLite cache engines
DEFAULT_CACHE_DIR = os.path.join(os.path.expanduser('~'), '.cache')
if not os.access(DEFAULT_CACHE_DIR, os.W_OK):
    # default cache dir is not writeable
    # therefore ask the os for a temp directory
    import tempfile
    DEFAULT_CACHE_DIR = tempfile.gettempdir()

API_KEY = os.getenv('NCBI_API_KEY', None)
if API_KEY:
    log.debug('NCBI_API_KEY found.')
else:
    log.warning('NCBI_API_KEY was not set.')

def get_process_log(filepath, loglevel=logging.INFO, name=PKGNAME+'.process'):
    "Sets up a file-based logger for process logging and returns its log object."
    log = logging.getLogger(name)
    log.setLevel(loglevel)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    fh = logging.FileHandler(filepath)
    fh.setFormatter(formatter)
    fh.setLevel(loglevel)
    log.addHandler(fh)
    return log

def get_data_log(filepath, name=PKGNAME+'.data'):
    "Sets up a file-based logger for data logging and returns its log object."
    datalog = logging.getLogger(name)
    datalog.setLevel(logging.DEBUG)
    datalog.propagate = False
    formatter = logging.Formatter('')
    fh = logging.FileHandler(filepath)
    fh.setFormatter(formatter)
    datalog.addHandler(fh)
    return datalog




from __future__ import absolute_import, unicode_literals

# Portlan Press Biochemical Society journals: mostly VIP
biochemsoc_format = 'http://{host}/content/{ja}/{a.volume}/{a.issue}/{a.first_page}.full.pdf'
biochemsoc_journals = {'Biochem J': {'host': 'www.biochemj.org', 'ja': 'ppbiochemj'},
                       'Clin Sci': {'host': 'www.clinsci.org', 'ja': 'ppclinsci'},
                       }



from metapub import UrlReverse

from metapub.urlreverse.methods import *

import unittest

# SAMPLE URLs (as yet unused in tests)
#
#https://www.researchgate.net/figure/279965765_fig3_Figure-2-Genotyping-the-CPVT1-RyR2-R420Q-mutation-and-CPVT2-CASQ2-D307H-mutation-A


# "fixtures"
pmid_samples = {
    'http://www.ncbi.nlm.nih.gov/pubmed/22253870': '22253870',
    'http://aac.asm.org/cgi/pmidlookup?view=long&pmid=7689822': '7689822',
    'http://science.sciencemag.org/lookup/pmidlookup?view=long&pmid=25143362.pdf': '25143362',
    }

vip_samples = {
    'http://www.jbc.org/content/266/17/10880.full.pdf': '2040605',
    'http://cardiovascres.oxfordjournals.org/content/75/1/69': '17449018',
    # 'http://rheumatology.oxfordjournals.org/content/48/6/704.2.full.pdf': '19304793',   # HTD
    'http://www.jbc.org/content/285/5/3076.full': '19920148',
    'http://www.bloodjournal.org/content/127/16/1967.full.pdf': '26932803', 
    'https://hmg.oxfordjournals.org/content/20/10/2048.full.pdf': '21378393',
    'http://jnci.oxfordjournals.org/content/100/4/277': '18270343',
    }

early_samples = {
    'http://cancerres.aacrjournals.org/content/early/2015/12/30/0008-5472.CAN-15-0295.full.pdf': '10.1158/0008-5472.CAN-15-0295',
    'http://ajcn.nutrition.org/content/early/2016/04/20/ajcn.115.123752.abstract': '10.3945/ajcn.115.123752',
    'http://www.mcponline.org/content/early/2016/04/25/mcp.O115.055467.full.pdf+html': '10.1074/mcp.O115.055467',
    'http://www.jbc.org/content/early/2001/01/08/jbc.M011064200.full.pdf': '10.1074/jbc.M011064200',
    'http://nar.oxfordjournals.org/content/early/2013/11/21/nar.gkt1163.full.pdf': '10.1093/nar/gkt1163',
    'http://jmg.bmj.com/content/early/2008/07/08/jmg.2008.058297': '10.1136/jmg.2008.058297',
    }

jstage_samples = { 
    'https://www.jstage.jst.go.jp/article/jat/16/3/16_E125/_pdf': '10.5551/jat.E125',
    'https://www.jstage.jst.go.jp/article/yoken/66/4/66_306/_pdf': '10.7883/yoken.66.306',
    'https://www.jstage.jst.go.jp/article/jvms1939/40/5/40_5_575/_pdf/': '10.1292/jvms1939.40.575',
    'https://www.jstage.jst.go.jp/article/analsci/21/12/21_12_1479/_article': '10.2116/analsci.21.1479',
    }
            
wiley_samples = {
    'http://onlinelibrary.wiley.com/doi/10.1111/j.1582-4934.2011.01476.x/full': '10.1111/j.1582-4934.2011.01476.x',
    'http://onlinelibrary.wiley.com/store/10.1002/(SICI)1097-0061(19980130)14:2<161::AID-YEA208>3.0.CO;2-Y/asset/208_ftp.pdf?v=1&t=ibqsvd4r&s=d74396a1e55e0a7b1bb08f297ce23c220d713d6f': '10.1002/(SICI)1097-0061(19980130)14:2<161::AID-YEA208>3.0.CO;2-Y',
    'http://onlinelibrary.wiley.com/doi/10.1002/humu.20182/pdf': '10.1002/humu.20182',
    }

karger_samples = {
    'http://www.karger.com/Article/PDF/322318': '10.1159/000322318',
    'https://www.karger.com/Article/Abstract/329047': '10.1159/000329047',
    'https://www.karger.com/Article/Abstract/83388': '10.1159/000083388',
    }

ahajournals_samples = {
    'http://circimaging.ahajournals.org/content/suppl/2013/04/02/CIRCIMAGING.112.000333.DC1/000333_Supplemental_Material.pdf': '10.1161/CIRCIMAGING.112.000333',
    }

sciencedirect_samples = {
    'http://www.sciencedirect.com/science/article/pii/S1386142500004509': '10.1016/S1386-1425(00)00450-9',
    'http://www.sciencedirect.com/science/article/pii/S131901641000071X': '10.1016/j.jsps.2010.07.008',
    }

cell_samples = {
    'http://www.cell.com/cell-reports/pdfExtended/S2211-1247(15)01030-X': '10.1016/j.celrep.2015.09.019',
    'http://www.cell.com/pdf/0092867480906212.pdf': '10.1016/0092-8674(80)90621-2',
    'http://www.cell.com/cancer-cell/pdf/S1535610806002844.pdf': '10.1016/j.ccr.2006.09.010',
    'http://www.cell.com/molecular-cell/abstract/S1097-2765(00)80321-4': '10.1016/S1097-2765(00)80321-4',
    'http://www.cell.com/biophysj/abstract/S0006-3495(15)01407-1': '10.1016/S0006-3495(15)01407-1',
    'http://www.cell.com/current-biology/fulltext/S0960-9822(16)30170-1': '10.1016/j.cub.2016.03.002',
    }

pmc_samples = {
    'http://europepmc.org/articles/pmc4103182': '24070655',
    'http://europepmc.org/backend/ptpmcrender.fcgi?accid=PMC3296117&blobtype=pdf': '21801150',
    'europepmc.org/articles/PMC360379/pdf/molcellb00133-0293.pdf': '1406642',
    'http://www.ncbi.nlm.nih.gov/pmc/articles/PMC4382744/': '25833843',
    'http://www.ncbi.nlm.nih.gov/pmc/articles/PMC3256213/': '22253870',
    'http://www.ncbi.nlm.nih.gov/pmc/articles/PMC3256213/pdf/pone.0030042.pdf': '22253870',
    }

jci_samples = {
    'https://www.jci.org/articles/view/32496': '10.1172/JCI32496',
    'http://www.jci.org/articles/view/37506/version/1/pdf/render': '10.1172/JCI37506',
    'http://www.jci.org/articles/view/18784': '10.1172/JCI18784',
    'https://www.jci.org/articles/view/31080/figure/1': '10.1172/JCI31080',
    }

spandidos_samples = {
    'http://www.spandidos-publications.com/or/30/2/553': '10.3892/or.2013.2535',
    'https://www.spandidos-publications.com/10.3892/or.2013.2535/abstract': '10.3892/or.2013.2535',
}

springer_samples = {
    'link.springer.com/article/10.1186/1471-2164-7-243': '10.1186/1471-2164-7-243',
    'http://link.springer.com/content/pdf/10.1007/s004390000422.pdf': '10.1007/s004390000422',
    }

bmj_samples = {
    'http://jmg.bmj.com/content/39/6/e31.full': '10.1136/jmg.39.6.e31',
    'http://www.bmj.com/content/353/bmj.i2195': '10.1136/bmj.i2195',
    'http://www.bmj.com/content/353/bmj.i2139': '10.1136/bmj.i2139',
    }

tough_cases = {
    'http://www.nature.com/pr/journal/v49/n2/full/pr200126a.html': '10.1203/00006450-200102000-00001',
    }

class TestUrlReverse(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_get_pmids_for_pmc_links(self):
        for url, pmid in pmc_samples.items():
            urlrev = UrlReverse(url, cachedir=None)
            assert urlrev.pmid == pmid

    def test_get_pmids_for_pmid_links(self):
        for url, pmid in pmid_samples.items():
            urlrev = UrlReverse(url, cachedir=None)
            assert urlrev.pmid == pmid

    def test_early_samples(self):
        for url, doi in early_samples.items():
            urlrev = UrlReverse(url, cachedir=None)
            assert urlrev.doi == doi

    def test_get_jstage_doi_from_link(self):
        for url, doi in jstage_samples.items():
            urlrev = UrlReverse(url, cachedir=None)
            assert doi == urlrev.doi

    def test_try_doi_methods_for_wiley_samples(self):
        for url, doi in wiley_samples.items():
            urlrev = UrlReverse(url, cachedir=None)
            assert urlrev.doi == doi

    def test_try_doi_methods_for_springer_samples(self):
        for url, doi in springer_samples.items():
            urlrev = UrlReverse(url, cachedir=None)
            assert urlrev.doi == doi

    def test_get_karger_doi_from_link(self):
        for url, doi in karger_samples.items():
            urlrev = UrlReverse(url, cachedir=None)
            assert doi == urlrev.doi

    def test_get_spandidos_doi_from_link(self):
        for url, doi in spandidos_samples.items():
            urlrev = UrlReverse(url, cachedir=None)
            assert doi == urlrev.doi

    def test_get_ahajournals_doi_from_link(self):
        for url, doi in ahajournals_samples.items():
            urlrev = UrlReverse(url, cachedir=None)
            assert doi == urlrev.doi

    def test_get_sciencedirect_doi_from_link(self):
        for url, doi in sciencedirect_samples.items():
            urlrev = UrlReverse(url, cachedir=None)
            assert doi == urlrev.doi

    def test_get_cell_doi_from_link(self):
        for url, doi in cell_samples.items():
            urlrev = UrlReverse(url, cachedir=None)
            assert doi == urlrev.doi

    def test_get_bmj_doi_from_link(self):
        for url, doi in bmj_samples.items():
            urlrev = UrlReverse(url, cachedir=None)
            assert doi == urlrev.doi

    # TODO: UrlReverse.suppled_info
    #def test_get_vip_url_with_supplied_title(self):
    #    url = 'http://www.clinsci.org/content/ppclinsci/130/11/871'
    #    expected_doi = '10.1042/CS20150777',
    #    urlrev = UrlReverse(url, title='High-fat diet increases O-GlcNAc levels in cerebral arteries')
    #    assert urlrev.doi == expected_doi

    def test_urlreverse_on_tough_cases(self):
        for url, doi in tough_cases.items():
            urlrev = UrlReverse(url)
            urlrev = UrlReverse(url, cachedir=None)
            assert urlrev.doi == doi

    def test_get_nature_doi_from_link(self):
        doi = get_nature_doi_from_link('http://www.nature.com/ejhg/journal/v22/n11/extref/ejhg201416x6.doc')
        assert doi == '10.1038/ejhg.2014.16'

        doi = get_nature_doi_from_link('http://www.nature.com/neuro/journal/v13/n11/abs/nn.2662.html')
        assert doi == '10.1038/nn.2662'

        doi = get_nature_doi_from_link('http://www.nature.com/gim/journal/vaop/ncurrent/extref/gim2014176x1.xls')
        assert doi == '10.1038/gim.2014.176'

        doi = get_nature_doi_from_link('http://www.nature.com/jhg/journal/v57/n3/pdf/jhg2011139a.pdf?origin=publication_detail')
        assert doi == '10.1038/jhg.2011.139'

        doi = get_nature_doi_from_link('http://www.nature.com/onc/journal/v26/n57/full/1210594a.html')
        assert doi == '10.1038/sj.onc.1210594'

        doi = get_nature_doi_from_link('http://www.nature.com/articles/cddis201475')
        assert doi == '10.1038/cddis.2014.75'

        doi = get_nature_doi_from_link('http://www.nature.com/articles/nature03404')
        assert doi == '10.1038/nature03404'

        doi = get_nature_doi_from_link('http://www.nature.com/articles/ng.2223')
        assert doi == '10.1038/ng.2223'

        doi = get_nature_doi_from_link('http://www.nature.com/pr/journal/v49/n1/abs/pr200116a.html')
        assert doi == '10.1203/00006450-200101000-00016'

        #doi = get_nature_doi_from_link('http://www.nature.com/pr/journal/v49/n1/full/pr200126a.pdf')
        #assert doi == '10.1203/00006450-200102000-00001'

    def test_get_biomedcentral_doi_from_link(self):
        doi = get_biomedcentral_doi_from_link('http://www.biomedcentral.com/content/pdf/bcr1282.pdf')
        assert doi == '10.1186/bcr1282'

        doi = get_biomedcentral_doi_from_link('http://www.biomedcentral.com/1471-2148/12/114')
        assert doi == '10.1186/1471-2148-12-114'

        doi = get_biomedcentral_doi_from_link('http://www.biomedcentral.com/content/supplementary/gb-2013-14-10-r108-S8.xlsx')
        assert doi == '10.1186/gb-2013-14-10-r108'

        doi = get_biomedcentral_doi_from_link('http://www.biomedcentral.com/1471-2164/15/707/table/T2')
        assert doi == '10.1186/1471-2164-15-707'




from __future__ import absolute_import, unicode_literals

"""metapub.MedGenFetcher -- tools to deal with NCBI's E-utilities interface to the MedGen db"""

from lxml import etree

from .eutils_common import get_eutils_client
from .cache_utils import get_cache_path 
from .exceptions import MetaPubError
from .medgenconcept import MedGenConcept
from .base import Borg, parse_elink_response
from .config import DEFAULT_EMAIL


class MedGenFetcher(Borg):
    """ MedGenFetcher (a Borg singleton object)

    An interaction layer for querying to return MedGenConcept objects.
    
    Currently available methods: eutils

    Basic Usage:

        fetch = MedGenFetcher()
    
    To specify a service method (more coming soon):

        fetch = MedGenFetcher('eutils')

    To return a MedGenConcept from a known UID:

        concept = fetch.concept_by_uid(known_UID)

    To return a list of UIDs relevant to a given term known in medgen:

        uids = fetch.uids_by_term(some_term)

    To get a medgen UID given a known Concept ID (cui):

        uid = fetch.uid_for_cui(known_cui)
    """

    _cache_filename = 'medgenfetcher.db'

    def __init__(self, method='eutils', cachedir='default'):
        self.method = method
        self._cache_path = None

        if method == 'eutils':
            self._cache_path = get_cache_path(cachedir, self._cache_filename)
            self.qs = get_eutils_client(self._cache_path)
            self.uids_by_term = self._eutils_uids_by_term
            self.concept_by_uid = self._eutils_concept_by_uid
            self.concept_by_cui = self._eutils_concept_by_cui
            self.uid_for_cui = self._eutils_uid_for_cui
            self.pubmeds_for_uid = self._eutils_pubmeds_for_uid
            self.pubmeds_for_cui = self._eutils_pubmeds_for_cui
        else:
            raise NotImplementedError('coming soon: fetch from local medgen via medgen-mysql.')

    def _eutils_uids_by_term(self, term):
        """ Wraps results of an medgen efetch term lookup, returning IDs of related MedGenConcepts.

        :param term: (str)
        :return uids: list of medgen uids
        :rtype: list
        """
        # http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?db=medgen&term=OCRL
        result = self.qs.esearch({'db': 'medgen', 'term': term})
        dom = etree.fromstring(result)
        uids = []
        idlist = dom.find('IdList')
        for item in idlist.findall('Id'):
            uids.append(item.text.strip())
        return uids
    
    def _eutils_uid_for_cui(self, cui):
        """ Given a ConceptID (cui), return a medgen ID.

        :param cui: (str)
        :return uid: (str)
        :rtype: str
        """
        # http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?db=medgen&term=C0000039
        if not cui.startswith('C'):
            raise MetaPubError('Invalid CUID: must start with C (e.g. C0000039)')

        result = self.qs.esearch({'db': 'medgen', 'term': cui})
        root = etree.fromstring(result).getroottree()
        try:
            uid = root.find('IdList').find('Id').text.strip()
        except AttributeError:
            raise MetaPubError('Invalid CUID: did not return MedGen id.')
        return uid

    def _eutils_concept_by_uid(self, uid):
        """ Returns MedGenConcept result of lookup of medgen uid.
        
        :param uid: (string or int) medgen uid
        :return: MedGenConcept or None
        :rtype: MedGenConcept object
        """
        # http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi?db=medgen&id=336867
        uid = str(uid)
        result = self.qs.esummary({'db': 'medgen', 'id': uid})
        return MedGenConcept(result)

    def _eutils_concept_by_cui(self, cui):
        """ Returns MedGenConcept result of lookup of CUI.
        
        :param cui: (string) Concept id (CUI)
        :return: MedGenConcept or None
        :rtype: MedGenConcept object
        """
        uid = self._eutils_uid_for_cui(cui)
        return self._eutils_concept_by_uid(uid)

    def _eutils_pubmeds_for_uid(self, uid):
        """ Returns list of pubmed IDs linked to this Medgen UID.

        :param uid: (str) Medgen UID
        :return: list of pubmed IDs (strings) or empty list
        :rtype: list
        """
        response = self.qs.elink({'dbfrom': 'medgen', 'id': uid, 'db': 'pubmed'})
        ids = parse_elink_response(response)
        return ids

    def _eutils_pubmeds_for_cui(self, cui):
        """ Given a ConceptID (cui), return a list of related pubmed article IDs.

        :param cui: (str) Medgen Concept ID (CUI)
        :return: list of pubmed IDs (strings) or empty list
        :rtype: list
        """
        uid = self._eutils_uid_for_cui(cui)
        return self._eutils_pubmeds_for_uid(uid)



import unittest

from metapub.convert import pmid2doi, PubMedArticle2doi, bookid2pmid
from metapub.crossref import TITLE_SIMILARITY_IDEAL_SCORE, TITLE_SIMILARITY_MIN_SCORE

import Levenshtein

pmid_with_doi_in_PMA = 25847151
pmid_with_doi_in_PMA_expected_doi = "10.1016/j.neulet.2015.04.001"

pmid_with_doi_from_CrossRef = 11228145
pmid_with_doi_from_CrossRef_expected_doi = '10.1126/science.1057766'

#TODO: find such an example?!?!?  the CrossRef rewire now works so well it's very uncommon to stump it! 
#pmid_with_unknown_doi = 19634325


# For testing expected Levenshtein distance on various similar academic titles. 
# see also https://www.intact-project.org/general/openapc/2018/01/29/doi-reverse-lookup/
title_pairs_equiv = [
    ('The phosphorylation of Hsp20 enhances its association with amyloid-\u03b2 to increase protection against neuronal cell death', 'The phosphorylation of Hsp20 enhances its association with amyloid-<beta> to increase protection against neuronal cell death'),
]

title_pairs_nonequiv = [
    ('Mutation of the TP53 gene and allelic imbalance at chromosome 17p13 in ductal carcinoma in situ.', 'Allelic expression imbalance of TP53 mutated and polymorphic alleles in head and neck tumors.'),   # .42
    ('Growth Hormones Links to Cancer.', 'CORRIGENDUM FOR "Growth Hormone\'s Links to Cancer".'),  # .77
    ('Anything but', 'anything goes'),  # .64 as-is, .72 after lowercasing!
]


NCBI_BOOKS = [
        {'book_id': 'NBK21248', 'pmid': None,},
        {'book_id': 'NBK26468', 'pmid': '20301790',},
        {'book_id': 'NBK1424', 'pmid': '20301597',},
        {'book_id': 'NBK201366', 'pmid': '24830047',},
]


class TestConversions(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_pmid2doi(self):
        # pmid2doi references PubMedArticle2doi, so let's consider that function
        # implicitly tested here.
        doi = pmid2doi(pmid_with_doi_in_PMA)
        assert doi == pmid_with_doi_in_PMA_expected_doi

        doi = pmid2doi(pmid_with_doi_from_CrossRef)
        assert doi == pmid_with_doi_from_CrossRef_expected_doi

        #doi = pmid2doi(pmid_with_unknown_doi)
        #assert doi is None

    def test_bookid2pmid(self):
        for item in NCBI_BOOKS:
            assert item['pmid'] == bookid2pmid(item['book_id'])

    def test_Levenshtein_pairs(self):

        for pair in title_pairs_equiv:
            assert Levenshtein.ratio(pair[0], pair[1]) > TITLE_SIMILARITY_IDEAL_SCORE

        for pair in title_pairs_nonequiv:
            assert Levenshtein.ratio(pair[0], pair[1]) < TITLE_SIMILARITY_MIN_SCORE




from __future__ import absolute_import, unicode_literals

# cell journals
# OLD cell_format = 'http://download.cell.com{ja}/pdf/PII{pii}.pdf'
cell_format = 'http://www.cell.com{ja}/pdf/{pii}.pdf'
cell_journals = {
    'Am J Hum Genet': {'ja': '/AJHG'},
    'Biophys J': {'ja': '/biophysj'},
    'Cancer Cell': {'ja': '/cancer-cell'},
    'Cell': {'ja': ''},
    'Cell Host Microbe': {'ja': '/cell-host-microbe'},
    'Cell Metab': {'ja': '/cell-metabolism'},
    'Cell Stem Cell': {'ja': '/cell-stem-cell'},
    'Chem Biol': {'ja': '/chemistry-biology'},
    'Curr Biol': {'ja': '/current-biology'},
    'Dev Cell': {'ja': '/developmental-cell'},
    'Immunity': {'ja': '/immunity'},
    'Mol Cell': {'ja': '/molecular-cell'},
    'Neuron': {'ja': '/neuron'},
    'Structure': {'ja': '/structure'},
    'Trends Mol Med': {'ja': '/trends'},
}



from __future__ import absolute_import, unicode_literals

# vip = Volume-Issue-Page format
#       URLs that have the same format except for the host name
#
# TODO: early release format

vip_format = 'http://{host}/content/{a.volume}/{a.issue}/{a.first_page}.full.pdf'

vip_journals = {
    'Ann Clin Biochem': {'host': 'acb.sagepub.com'},
    'Am J Clin Pathol': {'host': 'ajcp.ascpjournals.org'},
    'Am J Hypertens': {'host': 'ajh.oxfordjournals.org'},
    'Am J Physiol Cell Physiol': {'host': 'ajpcell.physiology.org'},
    'Am J Physiol Endocrinol Metab': {'host': 'ajpendo.physiology.org'},
    'Am J Physiol Gastrointest Liver Physiol': {'host': 'ajpgi.physiology.org'},
    'Am J Physiol Heart Circ Physiol': {'host': 'ajpheart.physiology.org'},
    'Am J Physiol Lung Cell Mol Physiol': {'host': 'ajplung.physiology.org'},
    'Am J Physiol Regul Integr Comp Physiol': {'host': 'ajpregu.physiology.org'},
    'Am J Physiol Renal Physiol': {'host': 'ajprenal.physiology.org'},
    # TODO: the rest of physiology.org. see: http://www.the-aps.org/mm/Publications/Journals
    'Angiology': {'host': 'ang.sagepub.com'}, 
    'Ann Oncol': {'host': 'annonc.oxfordjournals.org'},
    # TODO: backup_url: pmid lookup strategy, e.g.
    # http://aac.asm.org/cgi/pmidlookup?view=long&pmid=7689822
    'Anticancer Res': {'host': 'ar.iiarjournals.org'},
    'Antimicrob Agents Chemother': {'host': 'aac.asm.org'},
    'Appl Environ Microbiol': {'host': 'aem.asm.org'},
    'Arterioscler Thromb Vasc Biol': {'host': 'atvb.ahajournals.org'},
    'Assessment': {'host': 'asm.sagepub.com'},
    'Brain': {'host': 'brain.oxfordjournals.org'},
    'Breast Cancer Res': {'host': 'breast-cancer-research.com'},
    'Br J Anaesth': {'host': 'bja.oxfordjournals.org'},
    'Cancer Discov': {'host': 'cancerdiscovery.aacrjournals.org'},
    'Cancer Epidemiol Biomarkers Prev': {'host': 'cebp.aacrjournals.org'},
    'Cancer Res': {'host': 'cancerres.aacrjournals.org'},
    # TODO: backup_url: pmid lookup strategy, e.g.
    # http://www.cfp.ca/cgi/pmidlookup?view=long&pmid=19282532
    'Can Fam Physician': {'host': 'www.cfp.ca'},
    'Carcinogenesis': {'host': 'carcin.oxfordjournals.org'},
    'Cardiovasc Res': {'host': 'cardiovascres.oxfordjournals.org'},
    'Circulation': {'host': 'circ.ahajournals.org'},
    'Circ Arrhythm Electrophysiol': {'host': 'circep.ahajournals.org'},
    'Circ Cardiovasc Genet': {'host': 'circgenetics.ahajournals.org'},
    'Circ Res': {'host': 'circres.ahajournals.org'},
    'Clin Appl Thromb Hemost': {'host': 'cat.sagepub.com'},
    'Clin Cancer Res': {'host': 'clincancerres.aacrjournals.org'},
    'Clin Chem': {'host': 'clinchem.org'},
    'Clin Infect Dis': {'host': 'cid.oxfordjournals.org'},
    'Clin Microbiol Rev': {'host': 'cmr.asm.org'},
    'Clin Pediatr': {'host': 'cpj.sagepub.com'},
    'Clin Pediatr (Phila)': {'host': 'cpj.sagepub.com'},
    'Clin Vaccine Immunol': {'host': 'cvi.asm.org'},
    'Diabetes': {'host': 'diabetes.diabetesjournals.org'},
    'Diabetes Care': {'host': 'care.diabetesjournals.org'},
    'Drug Metab Dispos': {'host': 'dmd.aspetjournals.org'},
    # TODO: backup_url: pmid lookup strategy,
    # http://emboj.embopress.org/cgi/pmidlookup?view=long&pmid=9501081
    'EMBO J': {'host': 'emboj.embopress.org'},
    'Endocr Relat Cancer': {'host': 'erc.endocrinology-journals.org'},  #TODO this is now https://erc.bioscientifica.com/
    'Environ Entomol': {'host': 'ee.oxfordjournals.org'},
    'Eukaryot Cell': {'host': 'ec.asm.org'},
    'Eur Heart J': {'host': 'eurheartj.oxfordjournals.org'},
    'Eur J Endocrinol': {'host': 'eje-online.org'},
    'Eur Respir J': {'host': 'erj.ersjournals.com'},
    'FASEB J': {'host': 'fasebj.org'},
    'FEMS Microbiol Lett': {'host': 'femsle.oxfordjournals.org'},
    'Genome Announc': {'host': 'genomea.asm.org'},
    'Genome Biol': {'host': 'genomebiology.com'},
    'Genome Res': {'host': 'genome.cshlp.org'},
    'Genes Dev': {'host': 'genesdev.cshlp.org'},
    'Haematologica': {'host': 'haematologica.org'},
    'Hum Mol Genet': {'host': 'hmg.oxfordjournals.org'},
    'Hum Reprod': {'host': 'humrep.oxfordjournals.org'},
    'Hypertension': {'host': 'hyper.ahajournals.org'},
    'Infect Immun': {'host': 'iai.asm.org'},
    'Invest Ophthalmol Vis Sci': {'host': 'www.iovs.org'},
    'IOVS': {'host': 'iovs.org'},
    # TODO: backup_url: pmid lookup strategy,
    # http://jah.sagepub.com/cgi/pmidlookup?view=long&pmid=20056814
    'J Aging Health': {'host': 'jah.sagepub.com'},
    'J Am Soc Nephrol': {'host': 'jasn.asnjournals.org'},
    'J Antimicrob Chemother': {'host': 'jac.oxfordjournals.org'},
    'J Appl Physiol': {'host': 'jap.physiology.org'},
    # TODO: backup_url: pmid lookup strategy,
    # http://jb.asm.org/cgi/pmidlookup?view=long&pmid=7683021
    'J Bacteriol': {'host': 'jb.asm.org'},
    # TODO backup_url: pmid lookup strategy, e.g.
    # http://www.jbc.org/cgi/pmidlookup?view=long&pmid=14722075
    'J Biol Chem': {'host': 'www.jbc.org'},
    'J Bone Joint Surg Am': {'host': 'jbjs.org'},
    'J Cell Biol': {'host': 'jcb.rupress.org'},
    'J Cell Sci': {'host': 'jcs.biologists.org'},
    'J Child Neurol': {'host': 'jcn.sagepub.com'},
    'J Clin Oncol': {'host': 'jco.ascopubs.org'},
    'J Dent Res': {'host': 'jdr.sagepub.com'},
    #'J Endocrinol': {'host': 'joe.endocrinology-journals.org'},   #moved
    #'J Mol Endocrinol': {'host': 'jme.endocrinology-journals.org'}, # moved to endo.py
    'J Exp Med': {'host': 'jem.rupress.org'},
    'J Gerontol A Biol Sci Med Sci': {'host': 'biomedgerontology.oxfordjournals.org'},
    'J Hum Lact': {'host': 'jhl.sagepub.com'},
    'J Immunol': {'host': 'jimmunol.org'},
    'J Infect Dis': {'host': 'jid.oxfordjournals.org'},
    'J Lipid Res': {'host': 'www.jlr.org'},
    'J Clin Microbiol': {'host': 'jcm.asm.org'},
    'J Natl Cancer Inst': {'host': 'jnci.oxfordjournals.org'},
    'J Neurophysiol': {'host': 'jb.physiology.org'},
    'J Neurosci': {'host': 'jneurosci.org'},
    # TODO:  backup_url: pmid lookup strategy,
    # http://jn.nutrition.org/cgi/pmidlookup?view=long&pmid=10736367
    'J Nutr': {'host': 'jn.nutrition.org'},
    'J Pharmacol Exp Ther': {'host': 'jpet.aspetjournals.org'},
    'J Rheumatol': {'host': 'www.jrheum.org'},
    'J Renin Angiotensin Aldosterone Syst': {'host': 'jra.sagepub.com'},
    'J Virol': {'host': 'jvi.asm.org'},
    'Lupus': {'host': 'lup.sagepub.com'},
    'MBio': {'host': 'mbio.asm.org'},
    'mSystems': {'host': 'msystems.asm.org'},
    'mSphere': {'host': 'msphere.asm.org'},
    'Microbiol Mol Biol Rev': {'host': 'mmbr.asm.org'},
    'Mol Biol Cell': {'host': 'molbiolcell.org'},
    'Mol Cell Biol': {'host': 'mcb.asm.org'},
    'Mol Canc Therapeut': {'host': 'mct.aacrjournals.org'},
    'Mol Cancer Ther': {'host': 'mct.aacrjournals.org'},
    'Mol Hum Reprod': {'host': 'molehr.oxfordjournals.org'},
    'Mol Pharmacol': {'host': 'molpharm.aspetjournals.org'},
    'Mutagenesis': {'host': 'mutage.oxfordjournals.org'},
    'Neurology': {'host': 'neurology.org'},
    'Nephrol Dial Transplant': {'host': 'ndt.oxfordjournals.org'},
    'Nucleic Acids Res': {'host': 'nar.oxfordjournals.org'},
    'Orphanet J Rare Dis': {'host': 'ojrd.com'},
    'Pediatrics': {'host': 'pediatrics.aappublications.org'},
    # TODO: backup_url: pmid lookup strategy, e.g.
    # http://physiolgenomics.physiology.org/cgi/pmidlookup?view=long&pmid=15252189
    'Physiology (Bethesda)': {'host': 'physiologyonline.physiology.org'},
    'Physiol Genomics': {'host': 'physiolgenomics.physiology.org'},
    'Physiol Rep': {'host': 'physreports.physiology.org'},
    'Physiol Rev': {'host': 'physrev.physiology.org'},
    # TODO:  backup_url: pmid lookup strategy, e.g.
    # http://www.plantcell.org/cgi/pmidlookup?view=long&pmid=9501112
    'Plant Cell': {'host': 'www.plantcell.org'},
    'Plant Cell Physiol': {'host': 'pcp.oxfordjournals.org'},
    'Proc Natl Acad Sci USA': {'host': 'pnas.org'},
    'Protein Eng': {'host': 'peds.oxfordjournals.org'},
    'Protein Eng Des Sel': {'host': 'peds.oxfordjournals.org'},
    'QJM': {'host': 'qjmed.oxfordjournals.org'},
    'Radiat Res': {'host': 'jrr.oxfordjournals.org'},
    'Rheumatology (Oxford)': {'host': 'rheumatology.oxfordjournals.org'},
    'Science': {'host': 'sciencemag.org'},
    'Stroke': {'host': 'stroke.ahajournals.org'},

    'Acupunct Med': {'host': 'aim.bmj.com'},
    'Arch Dis Child': {'host': 'adc.bmj.com'},
    'Arch Dis Child Fetal Neonatal Ed': {'host': 'fn.bmj.com'},
    'Arch Dis Child Educ Pract Ed': {'host': 'ep.bmj.com'},
    'Ann Rheum Dis': {'host': 'ard.bmj.com'},
    'BMJ Innov': {'host': 'innovations.bmj.com'},
    'BMJ Open': {'host': 'bmjopen.bmj.com'},
    'BMJ Open Diabetes Res Care': {'host': 'drc.bmj.com'},
    'BMJ Open Resp Res': {'host': 'bmjopenrespres.bmj.com'},
    'BMJ Open Sport Exerc Med': {'host': 'bmjopensem.bmj.com'},
    'BMJ Qual Saf': {'host': 'qualitysafety.bmj.com'},
    'BMJ Qual Improv Report': {'host': 'qir.bmj.com'},
    'BMJ STEL': {'host': 'stel.bmj.com'},
    'BMJ Support Palliat Care': {'host': 'spcare.bmj.com'},
    'BMC Ophthalmol': {'host': 'bjo.bmj.com'},
    'Br J Sports Med': {'host': 'bjsm.bmj.com'},
    'Drug Ther Bull': {'host': 'dtb.bmj.com'},
    'Emerg Med J': {'host': 'emj.bmj.com'},
    'End Life J': {'host': 'eolj.bmj.com'},
    'ESMO Open': {'host': 'esmoopen.bmj.com'},
    'Eur J Hosp Pharm': {'host': 'ejhp.bmj.com'},
    'Evid Based Mental Health': {'host': 'ebmh.bmj.com'},
    'Evid Based Med': {'host': 'ebm.bmj.com'},
    'Evid Based Nurs': {'host': 'ebn.bmj.com'},
    'Frontline Gastroenterol': {'host': 'fg.bmj.com'},
    'Gut': {'host': 'gut.bmj.com'},
    'Heart': {'host': 'heart.bmj.com'},
    'Heart Asia': {'host': 'heartasia.bmj.com'},
    'Inj Prev': {'host': 'injuryprevention.bmj.com'},
    'In Pract': {'host': 'inpractice.bmj.com'},
    'J Clin Pathol': {'host': 'jcp.bmj.com'},
    'J Epidemiol Community Health': {'host': 'jech.bmj.com'},
    'J Fam Plann Reprod Health Care': {'host': 'jfprhc.bmj.com'},
    'J Investig Med': {'host': 'jim.bmj.com'},
    'J ISAKOS': {'host': 'jisakos.bmj.com'},
    'J Med Ethics': {'host': 'jme.bmj.com'},
    'J Med Genet': {'host': 'jmg.bmj.com'},
    'J Neurol Neurosurg Psychiatry': {'host': 'jnnp.bmj.com'},
    'J Neurointerv Surg': {'host': 'jnis.bmj.com'}, 
    'J R Army Med Corps': {'host': 'jramc.bmj.com'}, # not in Entrez journal list but has citation pmid 7602561
    'Lupus Sci Med': {'host': 'lupus.bmj.com'},
    'Med Humanities': {'host': 'mh.bmj.com'},
    'Occup Environ Med': {'host': 'oem.bmj.com'},
    'Open Heart': {'host': 'openheart.bmj.com'},
    'Pract Neurol': {'host': 'pn.bmj.com'},
    'RMD Open': {'host': 'rmdopen.bmj.com'},
    'Sex Transm Infect': {'host': 'sti.bmj.com'},
    'Vet Rec': {'host': 'veterinaryrecord.bmj.com'},
    'Vet Rec Case Rep': {'host': 'vetrecordcasereports.bmj.com'},
    'Vet Rec Open': {'host': 'vetrecordopen.bmj.com'},
    'Tob Control': {'host': 'tobaccocontrol.bmj.com'},
    'Postgrad Med J': {'host': 'pmj.bmj.com'},
    'Thorax': {'host': 'thorax.bmj.com'},
}

# volume-issue-page type URLs but with a nonstandard baseurl construction.
# e.g. Blood: http://www.bloodjournal.org/content/bloodjournal/79/10/2507.full.pdf
#      BMJ:   http://www.bmj.com/content/bmj/350/bmj.h3317.full.pdf
# Thorax:
# http://thorax.bmj.com/content/early/2015/06/23/thoraxjnl-2015-207199.full.pdf+html

# no trailing slash in baseurl (please)
vip_journals_nonstandard = {
    # TODO: backup_url: pmid lookup strategy, e.g.
    # http://www.bloodjournal.org/cgi/pmidlookup?view=long&pmid=1586703
    'Blood': 'http://www.bloodjournal.org/content/bloodjournal/{a.volume}/{a.issue}/{a.first_page}.full.pdf',
    'BMJ':   'http://www.bmj.com/content/bmj/{a.volume}/bmj.{a.first_page}.full.pdf',
}

# Non-VIP, not sure where to put yet
# 'BMJ Case Rep': {'host': 'casereports.bmj.com', 'example': 'http://casereports.bmj.com/content/2016/bcr-2015-214310'}
# 




__doc__ = "Utilities for cache file creation and management."
__author__ = "nthmost"

import os, logging
from datetime import datetime

# this is kinda ugh. we shouldn't be using an "internal" module, but the
# eutils package structure changed suddenly. Replace with bespoke...?
from eutils._internal.sqlitecache import SQLiteCache

from .config import PKGNAME, DEFAULT_CACHE_DIR
from .exceptions import MetaPubError


def datetime_to_timestamp(dt, epoch=datetime(1970,1,1)):
    """takes a python datetime object and converts it to a Unix timestamp.

    This is a non-timezone-aware function.

    :param dt: datetime to convert to timestamp
    :param epoch: datetime, option specification of start of epoch [default: 1/1/1970]
    :return: timestamp
    """
    td = dt - epoch
    return (td.microseconds + (td.seconds + td.days * 86400))


def get_cache_path(cachedir=DEFAULT_CACHE_DIR, filename='metapub-cache.db'):
    """ checks if cachedir exists; if not, tries to create it;
    raises MetaPubError if it can't be created.

        if cachedir is None, returns None.
    
        Default: DEFAULT_CACHE_DIR set in config.py (~/.cache)

        Supports expansion of user directory shortcut '~' to full path.

    :param cachedir: directory to store
    :param filename: name of cache file
    :return: path to SQLite DB file
    :raises MetaPubError
    """
    if cachedir is None:
        return None

    elif cachedir.find('~') > -1:
        cachedir = os.path.expanduser(cachedir)

    if _require_dir(cachedir):
        return os.path.join(cachedir, filename)
    else:
        raise MetaPubError('Could not create cache directory location %s' % cachedir)


def _require_dir(targetdir):
    if os.path.exists(targetdir):
        return True

    try:
        os.makedirs(targetdir)
        return True
    except OSError:
        return False

def cleanup_dir(cachedir):
    try:
        for item in os.listdir(cachedir):
            os.unlink(os.path.join(cachedir, item))
    except OSError:
        pass

    try:
        os.rmdir(cachedir)
    except OSError:
        pass


import glob
import os

from setuptools import setup, find_packages

setup(
    name='metapub',
    version='0.5.8',
    description='Pubmed / NCBI / eutils interaction library, handling the metadata of pubmed papers.',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    url='https://github.com/metapub/metapub',
    author='Naomi Most',
    maintainer='Naomi Most',
    author_email='naomi@nthmost.com',
    maintainer_email='naomi@nthmost.com',
    license='Apache 2.0',
    packages=find_packages(),
    entry_points={'console_scripts': [
        'pubmed_article = metapub.pubmedfetcher_cli:main',
        'convert = metapub.convert:main',
    ]
    },
    extras_require={
        'test': [
            'tox',
            'pytest',
        ],
    },
    install_requires=[
        'setuptools',
        'lxml',
        'lxml_html_clean',
        'requests',
        'eutils',
        'habanero',
        'tabulate',
        'cssselect',
        'unidecode',
        'docopt',
        'six',
        'coloredlogs',
        'python-Levenshtein',
    ],
)
